
from scipy import interpolate
import numpy as np

def goveqs_basis_disruption(input, t, M, notif_fac, i, prm, sel, agg):
     
    out = np.zeros(i['nx'])#len(input)
    invec = input[0:i['nstates']]

    if t<1980 :
        rHIV=0
    else:
        tempfn = interpolate.interp1d(np.array(range(len(prm['rHIV']))),prm['rHIV'],  fill_value="extrapolate")
        ti = t-1980
        rHIV = tempfn(ti) 
        #if ti < 0:
        #    rHIV = tempfn(0)
        #elif ti> len(prm['rHIV'])-1:
        #    rHIV = tempfn(len(prm['rHIV'])-1)
        #else:
        #    rHIV = tempfn(ti) 

    # Normalise by populations
    lam = M['lambda']*invec/sum(invec)
    #fac = interpolate.interp1(2019+(0:length(notif_fac))/12, [[1 1];notif_fac], t)
    
    yrs = [y/12+2019 for y in range(len(notif_fac))]
    fnfac = interpolate.interp1d(2019+np.array(range(len(notif_fac)+1))/12, np.concatenate([[1], np.concatenate(notif_fac)]), kind='next', fill_value="extrapolate")
    
    fac=fnfac(t) 

    #if t<2019:
    #    fac = fnfac(2019)
    #else:
    #    fac=fnfac(t) 

    # Pull them together
    allmat = M['lin'] + M['Dxlin']*fac  + rHIV*M['linHIV'] + lam[0]*M['nlin']['ds'] + lam[1]*M['nlin']['mdr']


    allmat = allmat.toarray()

    #invec_T = np.asmatrix(invec).T
    #out1 = np.matmul(allmat, invec_T)
    out1 = allmat@invec

    # Implement deaths
    #mortvec = [M['mortvec'][0]]*(i['nstates']+1)
    mortvec_sum = np.sum(M['mortvec'], 1)
    #morts = np.asmatrix(mortvec_sum*invec).T#[ mortvec_sum[idx]*inv  for idx, inv in enumerate(invec)]
    morts = mortvec_sum*invec
    out1 = out1 - morts

    #for idx in range(i['nstates']):
    #    out[idx] = out1[idx]
    #Changing the for loop since it is not efficient
    out[0:i['nstates']] = out1

    # Implement births
    births = sum(morts)
    out[i['U']['h0']['v0']] = out[i['U']['h0']['v0']]+births

    # Get the auxiliaries
    #sel_inci = (sel['inc']*allmat)
    #sel_noti = (sel['noti']*allmat)
    ##sel_mort = (sel['mort']*allmat)
    #sel_noti2 = (sel['noti2']*allmat)

    #sel_invec_inci = np.matmul(sel_inci, invec_T)
    #sel_invec_noti = np.matmul(sel_noti, invec_T)
    #sel_invec_noti2 = np.matmul(sel_noti2, invec_T) 

    #temp_arr = agg['inc']*sel_invec_inci
    #out[i['aux']['inc']]  = temp_arr.flatten()#np.matmul(agg['inc'], sel_invec_inci)
    #out[i['aux']['noti'][0]] = agg['noti'][0]*sel_invec_noti#np.matmul(agg['noti'][0], sel_invec_noti)
    #out[i['aux']['noti'][1]] = agg['noti'][1]*sel_invec_noti#np.matmul(agg['noti'][1], sel_invec_noti)
    
    #out[i['aux']['noti2']] = sum(sel['noti2']*sel_invec_noti)#sum(np.matmul(agg['noti2'], sel_invec_noti))

    #out[i['aux']['mort'][0]] = sum(M['mortvec'][:,1]*invec_T)
    #out[i['aux']['mort'][1]] = sum(M['mortvec'][:,2]*invec_T)
    
    #out = [o.sum() for o in out]

    out[i['aux']['inc']]  = agg['inc']@(sel['inc']*allmat)@invec
    out[i['aux']['noti']]  = agg['noti']@(sel['noti']*allmat)@invec
    out[i['aux']['noti2']] = sum((sel['noti2']*allmat)@invec)
    out[i['aux']['mort'][0]] = sum(M['mortvec'][:,1]*invec)
    out[i['aux']['mort'][1]] = sum(M['mortvec'][:,2]*invec)
            
    return out



from os import environ
import numpy as np

from AvenirCommon.Database import GB_get_db_json, GB_file_exists

from SpectrumCommon.Const.PJ.PJNTags import PJN_ISO3AlphaTag, PJN_FirstYearTag, PJN_FinalYearTag
from SpectrumCommon.Const.DP import DP_NaomiEstimateChoice, DP_NaomiFinalAgeCount
from SpectrumCommon.Const.TB import *
from SpectrumCommon.Modvars.TB.TBNotifDefs import *
from SpectrumCommon.Modvars.TB.TBModvars import TB_flatlineDB, TB_flatlinBurden, TB_getPopDefaults
from SpectrumCommon.Const.GB.GBVersions import (TB_WHO_DB_VERSION,
                                               TB_FORT_OUTPUT_VERSION)

def tb_apply_naomi_subnat(projection, subnat_region):
    countryID = projection[PJN_ISO3AlphaTag]
    
    connection = environ['AVENIR_SW_DEFAULT_DATA_CONNECTION']
    db =  GB_get_db_json(connection, 
                         'tuberculosis', 
                         f'countries\\{countryID}_{TB_WHO_DB_VERSION}.JSON')

    base_year = projection[PJN_FirstYearTag]
    final_year = projection[PJN_FinalYearTag]
    year_len = projection[PJN_FinalYearTag]-projection[PJN_FirstYearTag] + 1

    fort_slice = slice(max(2024, base_year)-TB_FirstHist_Year, min(TB_Fort_Final_Year, final_year) -TB_FirstHist_Year+1)
    tb_slice  = slice(max(2024-base_year, 0), min(TB_Fort_Final_Year, final_year)-base_year+1)

    
    first_idx = projection[PJN_FirstYearTag]-db['projections']['startYear']
    final_idx = projection[PJN_FinalYearTag]-db['projections']['startYear']

        # TB_HisIncidenceTag  : hist_inci,
        # TB_HisMortalityTag  : ,
    # flatline defaults 
    TB_flatlineDB(db, 'notification', final_idx)
    TB_flatlineDB(db, 'prevalencePct', final_idx)
    TB_flatlineDB(db, 'prevalence', final_idx)
    TB_flatlineDB(db, 'pop', final_idx)
    TB_flatlinBurden(db, 'e_inc_num', final_idx)
    TB_flatlinBurden(db, 'e_mort_num', final_idx)
    
    #subnat notification
    pop, pop_15_plus = TB_getPopDefaults(environ["AVENIR_SW_DEFAULT_DATA_CONNECTION"], countryID, subnat_region, projection[PJN_FirstYearTag], projection[PJN_FinalYearTag])
    DBName = f"subnationals\\{countryID}_subnat_V1.JSON"
    if GB_file_exists(environ["AVENIR_SW_DEFAULT_DATA_CONNECTION"], "tuberculosis", DBName):
        subnat = GB_get_db_json(environ["AVENIR_SW_DEFAULT_DATA_CONNECTION"], "tuberculosis", DBName)
        subnat = subnat[subnat_region]
        subnat_notif_ratio = subnat['notifProp']
        db['projections']['notification'] = [v*subnat['notifProp'] for v in db['projections']['notification'] ]
        db['projections']['prevalence'] = [v*subnat['notifProp'] for v in db['projections']['notification'] ]
        db['projections']['pop'] = [v*subnat['notifProp'] for v in db['projections']['notification'] ]
        db['projections']['e_inc_num'] = [v*subnat['notifProp'] for v in db['projections']['notification'] ]
        db['projections']['e_mort_num'] = [v*subnat['notifProp'] for v in db['projections']['notification'] ]
        for v in db['Noti_Distribution']:
            if '_perc' not in v:
                db['Noti_Distribution'][v] = db['Noti_Distribution'][v]*subnat['notifProp'] 
        demproj_subnat = GB_get_db_json(environ["AVENIR_SW_DEFAULT_DATA_CONNECTION"], 'demproj', f'subnationals\\{subnat_region}_V1.JSON')
        
        pop_ratio = np.array(demproj_subnat['population'][DP_NaomiEstimateChoice])[:,0:DP_NaomiFinalAgeCount].sum()/(pop[:, :, 0:, :, :].sum())
        pop_15_plus = np.array(demproj_subnat['population'][DP_NaomiEstimateChoice])[:,3:DP_NaomiFinalAgeCount].sum()
        pop = pop*pop_ratio
        pass
        projection[TB_SubnationalRatioTag]  = subnat_notif_ratio
        
        projection[TB_HisIncidenceTag] = [v*subnat_notif_ratio for v in db['burden']['e_inc_num'][first_idx:final_idx+1]]
        projection[TB_HisMortalityTag] = [v*subnat_notif_ratio for v in db['burden']['e_mort_num'][first_idx:final_idx+1]]
        
        
        notif_base_pct, notif_scale, c_ep, bac_pos_pct = create_notification_defaults_2022(countryID, db['projections']['notification'], db['Noti_Distribution'], 2024-db['projections']['startYear'])
        
        notif, notif_split, notif_pct, fort_outputs = create_notification_defaults(countryID, db['projections']['notification'], db['Noti_Distribution'], 
                                                                                   projection[PJN_FirstYearTag], 
                                                                                   first_idx, final_idx, year_len, notif_base_pct, tb_slice, fort_slice, 
                                                                                   TB_FORT_OUTPUT_VERSION, notif_scale, subnat_notif_ratio)
        
        projection[TB_HisNotificationTag] = notif
        projection[TB_TargNotificationTag] = notif
        projection[TB_ResNotificationTag] = np.array([notif, notif, notif*0])
        
        

        year_len = projection[PJN_FinalYearTag]-projection[PJN_FirstYearTag] + 1
        fort_slice = slice(max(2024, projection[PJN_FirstYearTag])-TB_FirstHist_Year, min(TB_Fort_Final_Year, projection[PJN_FinalYearTag]) -TB_FirstHist_Year+1)
        tb_slice  = slice(max(2024-projection[PJN_FirstYearTag], 0), min(TB_Fort_Final_Year, projection[PJN_FinalYearTag])-projection[PJN_FirstYearTag]+1)

        incidence = np.zeros(year_len)
        incidence[tb_slice] = [v*subnat_notif_ratio for v in fort_outputs['iMid'][fort_slice]]
        # flatline no longer needed, taking all of forts values
        # incidence[TB_Fort_Final_Year-base_year:] = incidence[TB_Fort_Final_Year-base_year] # flatline incidence from 2030 onwards
        
        mortality = np.zeros(year_len)
        mortality[tb_slice] = [v*subnat_notif_ratio for v in fort_outputs['mMid'][fort_slice]]

        # prevalence = calc_prevalence_from_notif_inci(notif, incidence)
        prevalence = np.zeros(year_len)
        prevalence[tb_slice] = [v*subnat_notif_ratio for v in fort_outputs['pMid'][fort_slice]]
        projection[TB_ResIncidenceTag ] =  np.array([incidence, incidence, incidence*0])
        projection[TB_ResMortalityTag ] =  np.array([mortality, mortality, mortality*0])
        projection[TB_ResPrevalenceTag] =  np.array([prevalence, prevalence, prevalence*0])
    return projection


import math as math

import SpectrumCommon.Const.GB.GBConst as gbc

import SpectrumCommon.Const.IC.ICModVarFields as icmvf
import SpectrumCommon.Const.IC.ICConst as icc

#######################################################################################################
#                                                                                                     #
#                                       Interventions                                                 #
#                                                                                                     #
#######################################################################################################


def get_IC_num_intervs(interv_records: list):
    return len(interv_records)


def get_IC_interv_rec_from_curr_ID(interv_records: list, curr_ID: int):
    return interv_records[curr_ID - 1]


def get_IC_interv_rec(interv_records: list, ID: str):
    interv_recs = [interv_rec for interv_rec in interv_records if interv_rec[icmvf.IC_INTERV_ID] == ID]
    return interv_recs[0] if len(interv_recs) > 0 else gbc.GB_NOT_FOUND

    # num_intervs = get_IC_num_intervs(interv_records)

    # stop = False
    # i = 0
    # while (i < num_intervs) and (not stop):
    #     interv_rec = interv_records[i]

    #     interv_ID = get_IC_interv_ID(interv_rec)

    #     if interv_ID == ID:
    #         stop = True
    #     else:
    #         i = i + 1

    # if stop:
    #     return interv_records[i]
    # else:
    #     return gbc.GB_NOT_FOUND


def get_IC_interv_idx(interv_records: list, ID: str):
    indices = [idx for (idx, interv_rec) in enumerate(interv_records) if interv_rec[icmvf.IC_INTERV_ID] == ID]
    return indices[0] if len(indices) > 0 else gbc.GB_NOT_FOUND

    # idx = gbc.GB_NOT_FOUND

    # i = 0
    # stop = False
    # num_intervs = get_IC_num_intervs(interv_records)

    # while (i < num_intervs) and (not stop):
    #     interv_rec = interv_records[i]

    #     interv_ID = get_IC_interv_ID(interv_rec)

    #     if interv_ID == ID:
    #         idx = i
    #         stop = True

    #     i += 1

    # return idx


###


def get_IC_interv_idx_and_rec(interv_records, ID: str):
    indices_and_recs = [
        (idx, interv_rec) for (idx, interv_rec) in enumerate(interv_records) if interv_rec[icmvf.IC_INTERV_ID] == ID
    ]
    return indices_and_recs[0] if len(indices_and_recs) > 0 else (gbc.GB_NOT_FOUND, {})


###


def get_IC_interv_ID(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_ID]


def set_IC_interv_ID(interv_rec: dict, value: str):
    interv_rec[icmvf.IC_INTERV_ID] = value


###


def get_IC_interv_name(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_NAME]


def set_IC_interv_name(interv_rec: dict, value: str):
    interv_rec[icmvf.IC_INTERV_NAME] = value


###


def get_IC_interv_string_constant(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_STR_CONST]


def set_IC_interv_string_constant(interv_rec: dict, value: str):
    interv_rec[icmvf.IC_INTERV_STR_CONST] = value


###


def get_IC_interv_active(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_ACTIVE]


def set_IC_interv_active(interv_rec: dict, value: bool):
    interv_rec[icmvf.IC_INTERV_ACTIVE] = value


###


def get_IC_impact_interv(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_IMPACT]


def set_IC_impact_interv(interv_rec: dict, value: bool):
    interv_rec[icmvf.IC_INTERV_IMPACT] = value


###


def get_IC_interv_custom(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_CUSTOM]


def set_IC_interv_custom(interv_rec: dict, value: bool):
    interv_rec[icmvf.IC_INTERV_CUSTOM] = value


###


def get_IC_interv_prog_area_ID(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_PROG_AREA_ID]


def set_IC_interv_prog_area_ID(interv_rec: dict, value: str):
    interv_rec[icmvf.IC_INTERV_PROG_AREA_ID] = value


###


def get_IC_interv_subgroup_ID(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_SUBGROUP_ID]


def set_IC_interv_subgroup_ID(interv_rec: dict, value: str):
    interv_rec[icmvf.IC_INTERV_SUBGROUP_ID] = value


###


def get_IC_interv_special_cases(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_SPECIAL_CASES]


def set_IC_interv_special_cases(interv_rec: dict, special_cases: list):
    interv_rec[icmvf.IC_INTERV_SPECIAL_CASES] = special_cases


def add_IC_interv_special_case(interv_rec: dict, mst_ID: str):
    interv_rec[icmvf.IC_INTERV_SPECIAL_CASES].append(mst_ID)
    return interv_rec


def remove_IC_interv_special_case(interv_rec: dict, mst_ID: str):
    interv_rec[icmvf.IC_INTERV_SPECIAL_CASES].remove(mst_ID)
    return interv_rec


###

# def get_IC_interv_PIN(interv_rec: dict, t: int):
#     return interv_rec[icmvf.IC_INTERV_PIN][t]

# def set_IC_interv_PIN(interv_rec: dict, t: int, value: float):
#     interv_rec[icmvf.IC_INTERV_PIN][t] = value

###


def get_IC_interv_targ_pop_direct_entry(interv_rec: dict, t: int):
    return interv_rec[icmvf.IC_INTERV_TARG_POP_REC][icmvf.IC_TARG_POP_DIRECT_ENTRY_VALUES][t]


def set_IC_interv_targ_pop_direct_entry(interv_rec: dict, t: int, value: float):
    interv_rec[icmvf.IC_INTERV_TARG_POP_REC][icmvf.IC_TARG_POP_DIRECT_ENTRY_VALUES][t] = value


###


def get_IC_interv_targ_pop_sex(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_TARG_POP_REC][icmvf.IC_TARG_POP_SEX]


def set_IC_interv_targ_pop_sex(interv_rec: dict, value: int):
    interv_rec[icmvf.IC_INTERV_TARG_POP_REC][icmvf.IC_TARG_POP_SEX] = value


###


def get_IC_interv_targ_pop_start_age(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_TARG_POP_REC][icmvf.IC_TARG_POP_START_AGE]


def set_IC_interv_targ_pop_start_age(interv_rec: dict, value: int):
    interv_rec[icmvf.IC_INTERV_TARG_POP_REC][icmvf.IC_TARG_POP_START_AGE] = value


###


def get_IC_interv_targ_pop_end_age(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_TARG_POP_REC][icmvf.IC_TARG_POP_END_AGE]


def set_IC_interv_targ_pop_end_age(interv_rec: dict, value: int):
    interv_rec[icmvf.IC_INTERV_TARG_POP_REC][icmvf.IC_TARG_POP_END_AGE] = value


###


def get_IC_interv_targ_pop_mst_ID(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_TARG_POP_REC][icmvf.IC_TARG_POP_MST_ID]


def set_IC_interv_targ_pop_mst_ID(interv_rec: dict, targ_pop_mstID: int):
    interv_rec[icmvf.IC_INTERV_TARG_POP_REC][icmvf.IC_TARG_POP_MST_ID] = targ_pop_mstID


###


def get_IC_interv_PIN_string_constant(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_PIN_STR_CONSTANT]


def set_IC_interv_PIN_string_constant(interv_rec: dict, value: str):
    interv_rec[icmvf.IC_INTERV_PIN_STR_CONSTANT] = value


###


def get_IC_interv_perc_deliv(interv_rec: dict, deliv_chan_IC_curr_ID: int, year_curr_ID: int):
    return interv_rec[icmvf.IC_INTERV_PERC_DELIV_THROUGH_CHAN][deliv_chan_IC_curr_ID - 1][year_curr_ID - 1]


def set_IC_interv_perc_deliv(interv_rec: dict, deliv_chan_IC_curr_ID: int, year_curr_ID: int, value: float):
    interv_rec[icmvf.IC_INTERV_PERC_DELIV_THROUGH_CHAN][deliv_chan_IC_curr_ID - 1][year_curr_ID - 1] = value


###


def get_IC_interv_PIN(interv_rec: dict, t: int):
    return interv_rec[icmvf.IC_INTERV_PIN][t]


def set_IC_interv_PIN(interv_rec: dict, t: int, value: float):
    interv_rec[icmvf.IC_INTERV_PIN][t] = value


###


def get_IC_interv_coverage(interv_rec: dict, t: int):
    return interv_rec[icmvf.IC_INTERV_COVERAGE][t]


def set_IC_interv_coverage(interv_rec: dict, t: int, value: float):
    interv_rec[icmvf.IC_INTERV_COVERAGE][t] = value


###


def get_IC_interv_PIN_editor_visited(interv_rec: dict):
    return interv_rec[icmvf.IC_INTERV_PIN_EDITOR_VISITED]


def set_IC_interv_PIN_editor_visited(interv_rec: dict, value: bool):
    interv_rec[icmvf.IC_INTERV_PIN_EDITOR_VISITED] = value


###

# These are small calculations that are part of the main calculations but also shown in the inputs

###

# def get_IC_interv_drug_cost_per_case(interv_rec: dict, deliv_chan_curr_ID: int, t: int):
#     return interv_rec[icmvf.IC_INTERV_DRUG_COST_PER_AVG_CASE][deliv_chan_curr_ID - 1][t]

# def set_IC_interv_drug_cost_per_case(interv_rec: dict, deliv_chan_curr_ID: int, t: int, value: float):
#     interv_rec[icmvf.IC_INTERV_DRUG_COST_PER_CASE][deliv_chan_curr_ID - 1][t] = value

# ###

# def get_IC_interv_supply_cost_per_case(interv_rec: dict, deliv_chan_curr_ID: int, t: int):
#     return interv_rec[icmvf.IC_INTERV_SUPPLY_COST_PER_CASE][deliv_chan_curr_ID - 1][t]

# def set_IC_interv_supply_cost_per_case(interv_rec: dict, deliv_chan_curr_ID: int, t: int, value: float):
#     interv_rec[icmvf.IC_INTERV_SUPPLY_COST_PER_CASE][deliv_chan_curr_ID - 1][t] = value

# ###

# def get_IC_interv_labor_cost_per_case(interv_rec: dict, deliv_chan_curr_ID: int, t: int):
#     return interv_rec[icmvf.IC_INTERV_LABOR_COST_PER_CASE][deliv_chan_curr_ID - 1][t]

# def set_IC_interv_labor_cost_per_case(interv_rec: dict, deliv_chan_curr_ID: int, t: int, value: float):
#     interv_rec[icmvf.IC_INTERV_LABOR_COST_PER_CASE][deliv_chan_curr_ID - 1][t] = value

# ###

# def get_IC_interv_min_per_avg_case_custom_staff(interv_rec: dict, deliv_chan_curr_ID: int, staff_type_curr_ID: int, t: int):
#     return interv_rec[icmvf.IC_INTERV_MIN_PER_AVG_CASE_CUSTOM_STAFF][deliv_chan_curr_ID - 1][staff_type_curr_ID - 1][t]

# def set_IC_interv_min_per_avg_case_custom_staff(interv_rec: dict, deliv_chan_curr_ID: int, staff_type_curr_ID: int, t: int, value: float):
#     interv_rec[icmvf.IC_INTERV_MIN_PER_AVG_CASE_CUSTOM_STAFF][deliv_chan_curr_ID - 1][staff_type_curr_ID - 1][t] = value

# ###

# def get_IC_interv_min_per_avg_case_default_staff(interv_rec: dict, deliv_chan_curr_ID: int, def_staff_type_curr_ID: int, t: int):
#     return interv_rec[icmvf.IC_INTERV_MIN_PER_AVG_CASE_DEFAULT_STAFF][deliv_chan_curr_ID - 1][def_staff_type_curr_ID - 1][t]

# def set_IC_interv_min_per_avg_case_default_staff(interv_rec: dict, deliv_chan_curr_ID: int, def_staff_type_curr_ID: int, t: int, value: float):
#     interv_rec[icmvf.IC_INTERV_MIN_PER_AVG_CASE_DEFAULT_STAFF][deliv_chan_curr_ID - 1][def_staff_type_curr_ID - 1][t] = value

#######################################################################################################
#                                                                                                     #
#                                       Drugs and supplies                                            #
#                                                                                                     #
#######################################################################################################


def get_IC_num_drugs_supplies(drug_supply_records: list):
    return len(drug_supply_records)


def get_IC_drug_supply_rec_from_curr_ID(drug_supply_records: list, curr_ID: int):
    return drug_supply_records[curr_ID - 1]


def get_IC_drug_supply_rec(drug_supply_records: list, ID: str):
    drug_supply_recs = [
        drug_supply_rec for drug_supply_rec in drug_supply_records if drug_supply_rec[icmvf.IC_DRUG_SUPPLY_ID] == ID
    ]
    return drug_supply_recs[0] if len(drug_supply_recs) > 0 else gbc.GB_NOT_FOUND

    # drug_supply_rec = None

    # ds = 0
    # stop = False
    # num_drug_supplies = get_IC_num_drugs_supplies(drug_supply_records)

    # while (ds < num_drug_supplies) and (not stop):

    #     drug_supply_ID = get_IC_drug_supply_ID(drug_supply_records[ds])

    #     if drug_supply_ID == ID:
    #         drug_supply_rec = drug_supply_records[ds]
    #         stop = True

    #     ds += 1

    # return drug_supply_rec


###


def get_IC_drug_supply_idx(drug_supply_records, ID: str):
    indices = [
        idx
        for (idx, drug_supply_rec) in enumerate(drug_supply_records)
        if drug_supply_rec[icmvf.IC_DRUG_SUPPLY_ID] == ID
    ]
    return indices[0] if len(indices) > 0 else gbc.GB_NOT_FOUND

    # idx = gbc.GB_NOT_FOUND

    # ds = 0
    # stop = False
    # num_drug_supplies = get_IC_num_drugs_supplies(drug_supply_records)

    # while (ds < num_drug_supplies) and (not stop):

    #     drug_supply_ID = get_IC_drug_supply_ID(drug_supply_records[ds])

    #     if drug_supply_ID == ID:
    #         idx = ds
    #         stop = True

    #     ds += 1

    # return idx


###


def get_IC_drug_supply_idx_and_rec(drug_supply_records, ID: str):
    indices_and_recs = [
        (idx, drug_supply_rec)
        for (idx, drug_supply_rec) in enumerate(drug_supply_records)
        if drug_supply_rec[icmvf.IC_DRUG_SUPPLY_ID] == ID
    ]
    return indices_and_recs[0] if len(indices_and_recs) > 0 else (gbc.GB_NOT_FOUND, {})


###


def get_IC_drug_supply_name(drug_supply_rec: dict):
    return drug_supply_rec[icmvf.IC_DRUG_SUPPLY_NAME]


def set_IC_drug_supply_name(drug_supply_rec: dict, value: str):
    drug_supply_rec[icmvf.IC_DRUG_SUPPLY_NAME] = value


###


def get_IC_drug_supply_string_constant(drug_supply_rec: dict):
    return drug_supply_rec[icmvf.IC_DRUG_SUPPLY_STR_CONST]


def set_IC_drug_supply_string_constant(drug_supply_rec: dict, value: str):
    drug_supply_rec[icmvf.IC_DRUG_SUPPLY_STR_CONST] = value


###


def get_IC_drug_supply_ID(drug_supply_rec: dict):
    return drug_supply_rec[icmvf.IC_DRUG_SUPPLY_ID]


def set_IC_drug_supply_ID(drug_supply_rec: dict, value: str):
    drug_supply_rec[icmvf.IC_DRUG_SUPPLY_ID] = value


###


def get_IC_drug_supply_active(drug_supply_rec: dict):
    return drug_supply_rec[icmvf.IC_DRUG_SUPPLY_ACTIVE]


def set_IC_drug_supply_active(drug_supply_rec: dict, value: bool):
    drug_supply_rec[icmvf.IC_DRUG_SUPPLY_ACTIVE] = value


###


def get_IC_drug_supply_custom(drug_supply_rec: dict):
    return drug_supply_rec[icmvf.IC_DRUG_SUPPLY_CUSTOM]


def set_IC_drug_supply_custom(drug_supply_rec: dict, value: bool):
    drug_supply_rec[icmvf.IC_DRUG_SUPPLY_CUSTOM] = value


###


def get_IC_drug_supply_type_mst_ID(drug_supply_rec: dict):
    return drug_supply_rec[icmvf.IC_DRUG_SUPPLY_TYPE_MST_ID]


def set_IC_drug_supply_type_mst_ID(drug_supply_rec: dict, value: int):
    drug_supply_rec[icmvf.IC_DRUG_SUPPLY_TYPE_MST_ID] = value


###


def get_IC_drug_supply_unit_cost(drug_supply_rec: dict, t: int):
    return drug_supply_rec[icmvf.IC_DRUG_SUPPLY_UNIT_COST][t]


def set_IC_drug_supply_unit_cost(drug_supply_rec: dict, t: int, value: float):
    drug_supply_rec[icmvf.IC_DRUG_SUPPLY_UNIT_COST][t] = value


###


def add_IC_drug_supply_applic_prog_area(drug_supply_rec: dict, prog_area_ID: str):
    drug_supply_rec[icmvf.IC_DRUG_SUPPLY_APPLIC_PROG_AREA_LIST].append(prog_area_ID)
    return drug_supply_rec


def remove_IC_drug_supply_applic_prog_area(drug_supply_rec: dict, prog_area_ID: str):
    drug_supply_rec[icmvf.IC_DRUG_SUPPLY_APPLIC_PROG_AREA_LIST].remove(prog_area_ID)
    return drug_supply_rec


def clear_IC_drug_supply_applic_prog_area(drug_supply_rec: dict):
    drug_supply_rec[icmvf.IC_DRUG_SUPPLY_APPLIC_PROG_AREA_LIST].clear()
    return drug_supply_rec


#######################################################################################################
#                                                                                                     #
#                                       Treatment input groups (TIGs)                                 #
#                                                                                                     #
#######################################################################################################


# Returns all TIG records for a particular intervention
def get_IC_TIG_records(TIG_records_all: list, interv_ID: str):
    # Reduce list to contain only records for the specified intervention.
    TIG_records_for_interv = [TIG_rec for TIG_rec in TIG_records_all if TIG_rec[icmvf.IC_TIG_INTERV_ID] == interv_ID]
    return TIG_records_for_interv


#


def add_IC_TIG_rec(TIG_records: list, value: dict):
    TIG_records.append(value)


###


def remove_IC_TIG_rec(TIG_records: list, interv_ID: str, deliv_chan_ID: str, treat_input_group_ID: str):
    TIG_recs = [
        TIG_rec
        for TIG_rec in TIG_records
        if not (
            (TIG_rec[icmvf.IC_TIG_INTERV_ID] == interv_ID)
            and (TIG_rec[icmvf.IC_TIG_DELIV_CHAN_ID] == deliv_chan_ID)
            and (TIG_rec[icmvf.IC_TIG_ID] == treat_input_group_ID)
        )
    ]
    return TIG_recs


###


def remove_IC_TIG_rec_by_name(TIG_records: list, interv_ID: str, deliv_chan_ID: str, name: str):
    TIG_recs = [
        TIG_rec
        for TIG_rec in TIG_records
        if not (
            (TIG_rec[icmvf.IC_TIG_INTERV_ID] == interv_ID)
            and (TIG_rec[icmvf.IC_TIG_DELIV_CHAN_ID] == deliv_chan_ID)
            and (TIG_rec[icmvf.IC_TIG_NAME] == name)
        )
    ]
    return TIG_recs


###


# Return all treatment input groups that are not associated with an intervention, or the ones associated with the intervention at the particular delivery
# channel specified. Used when copying treatment inputs so we can clear intervention treatment input groups at all channels except one.
def remove_IC_TIG_records_for_interv_except_deliv_chan(TIG_records: list, interv_ID: str, deliv_chan_ID: str):
    TIG_records_reduced = [
        TIG_rec
        for TIG_rec in TIG_records
        if (TIG_rec[icmvf.IC_TIG_INTERV_ID] != interv_ID) or (TIG_rec[icmvf.IC_TIG_DELIV_CHAN_ID] == deliv_chan_ID)
    ]
    return TIG_records_reduced


###


def remove_IC_TIG_records(TIG_records: list, interv_ID: str):
    TIG_records_reduced = [TIG_rec for TIG_rec in TIG_records if TIG_rec[icmvf.IC_TIG_INTERV_ID] != interv_ID]
    return TIG_records_reduced


###


def get_IC_TIG_rec(TIG_records: list, interv_ID: str, deliv_chan_ID: str, treat_input_group_ID: str):
    TIG_recs = [
        TIG_rec
        for TIG_rec in TIG_records
        if (
            (TIG_rec[icmvf.IC_TIG_INTERV_ID] == interv_ID)
            and (TIG_rec[icmvf.IC_TIG_DELIV_CHAN_ID] == deliv_chan_ID)
            and (TIG_rec[icmvf.IC_TIG_ID] == treat_input_group_ID)
        )
    ]
    return TIG_recs[0] if len(TIG_recs) > 0 else gbc.GB_NOT_FOUND

    # num_groups = len(TIG_records)
    # stop = False
    # i = 0

    # while (i < num_groups) and (not stop):
    #     TIG_rec_i = TIG_records[i]

    #     interv_ID_i = get_IC_TIG_interv_ID(TIG_rec_i)
    #     deliv_chan_ID_i = get_IC_TIG_deliv_chan_ID(TIG_rec_i)
    #     treat_input_group_ID_i = get_IC_TIG_ID(TIG_rec_i)

    #     if (interv_ID_i == interv_ID) and (deliv_chan_ID_i == deliv_chan_ID) and (treat_input_group_ID_i == treat_input_group_ID):
    #         stop = True
    #     else:
    #         i = i + 1

    # if stop:
    #     return TIG_records[i]
    # else:
    #     return gbc.GB_NOT_FOUND


###


def get_IC_TIG_rec_from_name(TIG_records: list, interv_ID: str, deliv_chan_ID: str, name: str):
    TIG_recs = [
        TIG_rec
        for TIG_rec in TIG_records
        if (
            (TIG_rec[icmvf.IC_TIG_INTERV_ID] == interv_ID)
            and (TIG_rec[icmvf.IC_TIG_DELIV_CHAN_ID] == deliv_chan_ID)
            and (TIG_rec[icmvf.IC_TIG_NAME] == name)
        )
    ]
    return TIG_recs[0] if len(TIG_recs) > 0 else gbc.GB_NOT_FOUND


###


def get_IC_TIG_records_for_interv_and_deliv_chan(TIG_records: list, interv_ID: str, deliv_chan_ID: str):
    TIG_records_reduced = [
        TIG_rec
        for TIG_rec in TIG_records
        if (TIG_rec[icmvf.IC_TIG_INTERV_ID] == interv_ID) and (TIG_rec[icmvf.IC_TIG_DELIV_CHAN_ID] == deliv_chan_ID)
    ]
    return TIG_records_reduced


###


def get_IC_TIG_idx(TIG_records: list, interv_ID: str, deliv_chan_ID: str, treat_input_group_ID: str):
    indices = [
        idx
        for (idx, TIG_rec) in enumerate(TIG_records)
        if (TIG_rec[icmvf.IC_TIG_INTERV_ID] == interv_ID)
        and (TIG_rec[icmvf.IC_TIG_DELIV_CHAN_ID] == deliv_chan_ID)
        and (TIG_rec[icmvf.IC_TIG_ID] == treat_input_group_ID)
    ]
    return indices[0] if len(indices) > 0 else gbc.GB_NOT_FOUND

    # idx = gbc.GB_NOT_FOUND

    # i = 0
    # stop = False
    # num_groups = len(TIG_records)

    # while (i < num_groups) and (not stop):
    #     TIG_rec_i = TIG_records[i]

    #     interv_ID_i = get_IC_TIG_interv_ID(TIG_rec_i)
    #     deliv_chan_ID_i = get_IC_TIG_deliv_chan_ID(TIG_rec_i)
    #     treat_input_group_ID_i = get_IC_TIG_ID(TIG_rec_i)

    #     if (interv_ID_i == interv_ID) and (deliv_chan_ID_i == deliv_chan_ID) and (treat_input_group_ID_i == treat_input_group_ID):
    #         idx = i
    #         stop = True

    #     i = i + 1

    # return idx


###


def get_IC_TIG_name(treat_input_group_rec: dict):
    return treat_input_group_rec[icmvf.IC_TIG_NAME]


def set_IC_TIG_name(treat_input_group_rec: dict, value: str):
    treat_input_group_rec[icmvf.IC_TIG_NAME] = value
    # Unconventional, but allows us to use in list comprehensions to set values.
    return treat_input_group_rec


def set_IC_TIG_names_for_interv_and_deliv_chan(
    TIG_records: list, interv_ID: str, deliv_chan_ID: str, old_name: str, new_name: str
):
    for TIG_rec in TIG_records:
        if (
            TIG_rec[icmvf.IC_TIG_INTERV_ID] == interv_ID
            and TIG_rec[icmvf.IC_TIG_DELIV_CHAN_ID] == deliv_chan_ID
            and TIG_rec[icmvf.IC_TIG_NAME] == old_name
        ):
            set_IC_TIG_name(TIG_rec, new_name)
    return TIG_records


###


def get_IC_TIG_string_constant(treat_input_group_rec: dict):
    return treat_input_group_rec[icmvf.IC_TIG_STR_CONST]


def set_IC_TIG_string_constant(treat_input_group_rec: dict, value: str):
    treat_input_group_rec[icmvf.IC_TIG_STR_CONST] = value


###


def get_IC_TIG_interv_ID(treat_input_group_rec: dict):
    return treat_input_group_rec[icmvf.IC_TIG_INTERV_ID]


def set_IC_TIG_interv_ID(treat_input_group_rec: dict, value: str):
    treat_input_group_rec[icmvf.IC_TIG_INTERV_ID] = value


###


def get_IC_TIG_deliv_chan_ID(treat_input_group_rec: dict):
    return treat_input_group_rec[icmvf.IC_TIG_DELIV_CHAN_ID]


def set_IC_TIG_deliv_chan_ID(treat_input_group_rec: dict, value: str):
    treat_input_group_rec[icmvf.IC_TIG_DELIV_CHAN_ID] = value


###


def get_IC_TIG_ID(treat_input_group_rec: dict):
    return treat_input_group_rec[icmvf.IC_TIG_ID]


def set_IC_TIG_ID(treat_input_group_rec: dict, value: str):
    treat_input_group_rec[icmvf.IC_TIG_ID] = value


###

#######################################################################################################
#                                                                                                     #
#                                       Treatment inputs (TI) - all                                   #
#                                                                                                     #
#######################################################################################################


def get_IC_TI_rec(TI_records: list, interv_ID: str, deliv_chan_ID: str, TIG_ID: str, TI_ID: str):
    TI_recs = [
        TI_rec
        for TI_rec in TI_records
        if (
            (TI_rec[icmvf.IC_TI_INTERV_ID] == interv_ID)
            and (TI_rec[icmvf.IC_TI_DELIV_CHAN_ID] == deliv_chan_ID)
            and (TI_rec[icmvf.IC_TI_GROUP_ID] == TIG_ID)
            and (TI_rec[icmvf.IC_TI_ID] == TI_ID)
        )
    ]
    return TI_recs[0] if len(TI_recs) > 0 else gbc.GB_NOT_FOUND


def get_IC_TI_records_for_interv(TI_records: list, interv_ID: str):
    # Reduce list to contain only records for the specified intervention.
    TI_records_for_interv = [TI_rec for TI_rec in TI_records if TI_rec[icmvf.IC_TI_INTERV_ID] == interv_ID]
    return TI_records_for_interv


###


def get_IC_TI_records_for_deliv_chan(TI_records: list, deliv_chan_ID: str):
    # Reduce list to contain only records for the specified delivery channel.
    TI_records_for_deliv_chan = [TI_rec for TI_rec in TI_records if TI_rec[icmvf.IC_TI_DELIV_CHAN_ID] == deliv_chan_ID]
    return TI_records_for_deliv_chan


###


def get_IC_TI_records_for_interv_and_deliv_chan(TI_records: list, interv_ID: str, deliv_chan_ID: str):
    TI_records_reduced = [
        TI_rec
        for TI_rec in TI_records
        if (TI_rec[icmvf.IC_TI_INTERV_ID] == interv_ID) and (TI_rec[icmvf.IC_TI_DELIV_CHAN_ID] == deliv_chan_ID)
    ]
    return TI_records_reduced


###


# Useful for getting all treatment inputs that are the same for multiple delivery channels.
def get_IC_TI_records_for_interv_group_input(TI_records: list, TI_type_mst_ID: int, TI_rec_match: dict):
    TI_records_reduced = [
        TI_rec
        for TI_rec in TI_records
        if (
            (TI_rec[icmvf.IC_TI_INTERV_ID] == TI_rec_match[icmvf.IC_TI_INTERV_ID])
            and (TI_rec[icmvf.IC_TI_GROUP_ID] == TI_rec_match[icmvf.IC_TI_GROUP_ID])
            and (TI_rec[icmvf.IC_TI_ID] == TI_rec_match[icmvf.IC_TI_ID])
            and (math.isclose(TI_rec[icmvf.IC_TI_PERC_APPLIED][0], TI_rec_match[icmvf.IC_TI_PERC_APPLIED][0]))
            and (TI_rec[icmvf.IC_TI_NOTE] == TI_rec_match[icmvf.IC_TI_NOTE])
            and (
                (
                    (TI_type_mst_ID == icc.IC_TI_DRUGS_SUPPLIES_MST_ID)
                    and (
                        math.isclose(
                            TI_rec[icmvf.IC_TI_DRUG_SUPPLY_TIMES_PER_DAY],
                            TI_rec_match[icmvf.IC_TI_DRUG_SUPPLY_TIMES_PER_DAY],
                        )
                    )
                    and (
                        math.isclose(
                            TI_rec[icmvf.IC_TI_DRUG_SUPPLY_DAYS_PER_CASE],
                            TI_rec_match[icmvf.IC_TI_DRUG_SUPPLY_DAYS_PER_CASE],
                        )
                    )
                    and (
                        math.isclose(
                            TI_rec[icmvf.IC_TI_DRUG_SUPPLY_NUMBER], TI_rec_match[icmvf.IC_TI_DRUG_SUPPLY_NUMBER]
                        )
                    )
                )
                or (
                    (TI_type_mst_ID == icc.IC_TI_MED_PERSONNEL_TIME_MST_ID)
                    and (
                        math.isclose(
                            TI_rec[icmvf.IC_TI_MED_PERSONNEL_NUM_OF_DAYS],
                            TI_rec_match[icmvf.IC_TI_MED_PERSONNEL_NUM_OF_DAYS],
                        )
                    )
                    and (
                        math.isclose(
                            TI_rec[icmvf.IC_TI_MED_PERSONNEL_MINUTES], TI_rec_match[icmvf.IC_TI_MED_PERSONNEL_MINUTES]
                        )
                    )
                )
                or (
                    (TI_type_mst_ID == icc.IC_TI_OUTPAT_VISITS_INPAT_DAYS_MST_ID)
                    and (
                        math.isclose(
                            TI_rec[icmvf.IC_TI_OUTPAT_INPAT_UNITS_PER_CASE],
                            TI_rec_match[icmvf.IC_TI_OUTPAT_INPAT_UNITS_PER_CASE],
                        )
                    )
                )
            )
        )
    ]
    return TI_records_reduced


###


def add_IC_TI_rec(TI_records: list, value: dict):
    TI_records.append(value)


##


# Removes a single treatment input record.
def remove_IC_TI_rec(TI_records_all: list, interv_ID: str, deliv_chan_ID: str, TIG_ID: str, TI_ID: str):
    TI_recs = [
        TI_rec
        for TI_rec in TI_records_all
        if not (
            (TI_rec[icmvf.IC_TI_INTERV_ID] == interv_ID)
            and (TI_rec[icmvf.IC_TI_DELIV_CHAN_ID] == deliv_chan_ID)
            and (TI_rec[icmvf.IC_TI_GROUP_ID] == TIG_ID)
            and (TI_rec[icmvf.IC_TI_ID] == TI_ID)
        )
    ]
    return TI_recs

    # idx = gbc.GB_NOT_FOUND

    # i = 0
    # stop = False
    # num_treat_inputs = len(TI_records_all)

    # while (i < num_treat_inputs) and (not stop):
    #     TI_rec_i = TI_records_all[i]

    #     interv_ID_i = get_IC_TI_interv_ID(TI_rec_i)
    #     deliv_chan_ID_i = get_IC_TI_deliv_chan_ID(TI_rec_i)
    #     treat_input_group_ID_i = get_IC_TI_group_ID(TI_rec_i)
    #     treat_input_ID_i = get_IC_TI_ID(TI_rec_i)

    #     if (interv_ID_i == interv_ID) and (deliv_chan_ID_i == deliv_chan_ID) and (treat_input_group_ID_i == treat_input_group_ID) and (treat_input_ID_i == treat_input_ID):
    #         del TI_records_all[i]
    #         stop = True

    #     i = i + 1

    # return idx


###


# Removes all treatment input records associated with a particular intervention
def remove_IC_TI_records_for_interv(TI_records: list, interv_ID: str):
    TI_records_reduced = [TI_rec for TI_rec in TI_records if TI_rec[icmvf.IC_TI_INTERV_ID] != interv_ID]
    return TI_records_reduced


# Return all treatment inputs that are not associated with an intervention, or the ones associated with the intervention at the particular delivery
# channel specified. Used when copying treatment inputs so we can clear intervention treatment inputs at all channels except one.
def remove_IC_TI_records_for_interv_except_deliv_chan(TI_records: list, interv_ID: str, deliv_chan_ID: str):
    TI_records_reduced = [
        TI_rec
        for TI_rec in TI_records
        if (TI_rec[icmvf.IC_TI_INTERV_ID] != interv_ID) or (TI_rec[icmvf.IC_TI_DELIV_CHAN_ID] == deliv_chan_ID)
    ]
    return TI_records_reduced


# Removes all treatment input records in a particular group.
def remove_IC_TI_records_for_group(TI_records_all: list, interv_ID: str, deliv_chan_ID: str, TIG_ID: str):
    TI_recs = [
        TI_rec
        for TI_rec in TI_records_all
        if not (
            (TI_rec[icmvf.IC_TI_INTERV_ID] == interv_ID)
            and (TI_rec[icmvf.IC_TI_DELIV_CHAN_ID] == deliv_chan_ID)
            and (TI_rec[icmvf.IC_TI_GROUP_ID] == TIG_ID)
        )
    ]
    return TI_recs


###


def get_IC_TI_interv_ID(treat_input_rec: dict):
    return treat_input_rec[icmvf.IC_TI_INTERV_ID]


def set_IC_TI_interv_ID(treat_input_rec: dict, value: str):
    treat_input_rec[icmvf.IC_TI_INTERV_ID] = value


###


def get_IC_TI_deliv_chan_ID(treat_input_rec: dict):
    return treat_input_rec[icmvf.IC_TI_DELIV_CHAN_ID]


def set_IC_TI_deliv_chan_ID(treat_input_rec: dict, value: str):
    treat_input_rec[icmvf.IC_TI_DELIV_CHAN_ID] = value


###


def get_IC_TI_group_name(treat_input_rec: dict):
    return treat_input_rec[icmvf.IC_TIG_NAME]


def set_IC_TI_group_name(treat_input_rec: dict, value: str):
    treat_input_rec[icmvf.IC_TIG_NAME] = value


###


def get_IC_TI_group_string_constant(treat_input_rec: dict):
    return treat_input_rec[icmvf.IC_TIG_STR_CONST]


def set_IC_TI_group_string_constant(treat_input_rec: dict, value: str):
    treat_input_rec[icmvf.IC_TIG_STR_CONST] = value


###


def get_IC_TI_group_ID(treat_input_rec: dict):
    return treat_input_rec[icmvf.IC_TIG_ID]


def set_IC_TI_group_ID(treat_input_rec: dict, value: str):
    treat_input_rec[icmvf.IC_TIG_ID] = value


###


def get_IC_TI_ID(treat_input_rec: dict):
    return treat_input_rec[icmvf.IC_TI_ID]


def set_IC_TI_ID(treat_input_rec: dict, value: str):
    treat_input_rec[icmvf.IC_TI_ID] = value


###


def get_IC_TI_perc_applied(treat_input_rec: dict, t: int):
    return treat_input_rec[icmvf.IC_TI_PERC_APPLIED][t]


def set_IC_TI_perc_applied(treat_input_rec: dict, t: int, value: float):
    treat_input_rec[icmvf.IC_TI_PERC_APPLIED][t] = value


###


def get_IC_TI_note_string_constant(treat_input_rec: dict):
    return treat_input_rec[icmvf.IC_TI_NOTE_STR_CONST]


def set_IC_TI_note_string_constant(treat_input_rec: dict, value: str):
    treat_input_rec[icmvf.IC_TI_NOTE_STR_CONST] = value


###


def get_IC_TI_note(treat_input_rec: dict):
    return treat_input_rec[icmvf.IC_TI_NOTE]


def set_IC_TI_note(treat_input_rec: dict, value: str):
    treat_input_rec[icmvf.IC_TI_NOTE] = value


#######################################################################################################
#                                                                                                     #
#                                  Treatment inputs (TI) - drugs and supplies only                    #
#                                                                                                     #
#######################################################################################################


def get_IC_TI_drug_supply_times_per_day(TI_drug_supply_rec: dict):
    return TI_drug_supply_rec[icmvf.IC_TI_DRUG_SUPPLY_TIMES_PER_DAY]


def set_IC_TI_drug_supply_times_per_day(TI_drug_supply_rec: dict, value: int):
    TI_drug_supply_rec[icmvf.IC_TI_DRUG_SUPPLY_TIMES_PER_DAY] = value


###


def get_IC_TI_drug_supply_days_per_case(TI_drug_supply_rec: dict):
    return TI_drug_supply_rec[icmvf.IC_TI_DRUG_SUPPLY_DAYS_PER_CASE]


def set_IC_TI_drug_supply_days_per_case(TI_drug_supply_rec: dict, value: int):
    TI_drug_supply_rec[icmvf.IC_TI_DRUG_SUPPLY_DAYS_PER_CASE] = value


###


def get_IC_TI_drug_supply_number(TI_drug_supply_rec: dict):
    return TI_drug_supply_rec[icmvf.IC_TI_DRUG_SUPPLY_NUMBER]


def set_IC_TI_drug_supply_number(TI_drug_supply_rec: dict, value: int):
    TI_drug_supply_rec[icmvf.IC_TI_DRUG_SUPPLY_NUMBER] = value


#######################################################################################################
#                                                                                                     #
#                          Treatment inputs - medical personnel time only                             #
#                                                                                                     #
#######################################################################################################


def get_IC_TI_med_personnel_num_days(TI_med_personnel_rec: dict):
    return TI_med_personnel_rec[icmvf.IC_TI_MED_PERSONNEL_NUM_OF_DAYS]


def set_IC_TI_med_personnel_num_days(TI_med_personnel_rec: dict, value: int):
    TI_med_personnel_rec[icmvf.IC_TI_MED_PERSONNEL_NUM_OF_DAYS] = value


###


def get_IC_TI_med_personnel_minutes(TI_med_personnel_rec: dict):
    return TI_med_personnel_rec[icmvf.IC_TI_MED_PERSONNEL_MINUTES]


def set_IC_TI_med_personnel_minutes(TI_med_personnel_rec: dict, value: float):
    TI_med_personnel_rec[icmvf.IC_TI_MED_PERSONNEL_MINUTES] = value


###


def get_IC_TI_med_personnel_custom(TI_med_personnel_rec: dict):
    return TI_med_personnel_rec[icmvf.IC_TI_MED_PERSONNEL_CUSTOM_STAFF_TYPE]


def set_IC_TI_med_personnel_custom(TI_med_personnel_rec: dict, value: bool):
    TI_med_personnel_rec[icmvf.IC_TI_MED_PERSONNEL_CUSTOM_STAFF_TYPE] = value


#######################################################################################################
#                                                                                                     #
#                           Treatment inputs - outpatient days and inpatient visits only              #
#                                                                                                     #
#######################################################################################################


def get_IC_TI_outpat_inpat_units_per_case(TI_outpat_inpat_rec: dict):
    return TI_outpat_inpat_rec[icmvf.IC_TI_OUTPAT_INPAT_UNITS_PER_CASE]


def set_IC_TI_outpat_inpat_units_per_case(TI_outpat_inpat_rec: dict, value: float):
    TI_outpat_inpat_rec[icmvf.IC_TI_OUTPAT_INPAT_UNITS_PER_CASE] = value


#######################################################################################################
#                                                                                                     #
#                                       Target populations                                            #
#                                                                                                     #
#######################################################################################################


def get_IC_num_def_targ_pops(targ_pop_records: list):
    return len(targ_pop_records)


###


def get_IC_def_targ_pop_rec(targ_pop_records: list = [], mst_ID: int | None = None):
    targ_pop_recs = [
        targ_pop_rec for targ_pop_rec in targ_pop_records if targ_pop_rec[icmvf.IC_TARG_POP_DEF_MST_ID] == mst_ID
    ]
    return targ_pop_recs[0] if len(targ_pop_recs) > 0 else gbc.GB_NOT_FOUND

    # num_targ_pops = get_num_targ_pops(targ_pop_records)

    # stop = False
    # i = 0
    # while (i < num_targ_pops - 1) and (not stop):
    #     targ_pop_rec = targ_pop_records[i]

    #     mst_ID_i = get_IC_targ_pop_mst_ID(targ_pop_rec)

    #     if mst_ID_i == mst_ID:
    #         stop = True
    #     else:
    #         i = i + 1

    # if stop:
    #     return targ_pop_records[i]
    # else:
    #     return gbc.GB_NOT_FOUND


###


def get_IC_def_targ_pop_mst_ID(targ_pop_rec: dict):
    return targ_pop_rec[icmvf.IC_TARG_POP_DEF_MST_ID]


def set_IC_def_targ_pop_mst_ID(targ_pop_rec: dict, value: int):
    targ_pop_rec[icmvf.IC_TARG_POP_DEF_MST_ID] = value


###


def get_IC_def_targ_pop_string_constant(targ_pop_rec: dict):
    return targ_pop_rec[icmvf.IC_TARG_POP_DEF_STR_CONST]


def set_IC_def_targ_pop_string_constant(targ_pop_rec: dict, value: str):
    targ_pop_rec[icmvf.IC_TARG_POP_DEF_STR_CONST] = value


###


def get_IC_def_targ_pop_name(targ_pop_rec: dict):
    return targ_pop_rec[icmvf.IC_TARG_POP_DEF_NAME]


def set_IC_def_targ_pop_name(targ_pop_rec: dict, value: str):
    targ_pop_rec[icmvf.IC_TARG_POP_DEF_NAME] = value


###


def add_IC_def_targ_pop_applic_prog_area(targ_pop_rec: dict, prog_area_ID: str):
    targ_pop_rec[icmvf.IC_TARG_POP_DEF_APPLIC_PROG_AREA_IDS].append(prog_area_ID)
    return targ_pop_rec


def remove_IC_def_targ_pop_applic_prog_area(targ_pop_rec: dict, prog_area_ID: str):
    targ_pop_rec[icmvf.IC_TARG_POP_DEF_APPLIC_PROG_AREA_IDS].remove(prog_area_ID)
    return targ_pop_rec


def get_IC_def_targ_pop_applic_prog_area(targ_pop_rec: dict, prog_area_ID: str):
    return prog_area_ID in targ_pop_rec[icmvf.IC_TARG_POP_DEF_APPLIC_PROG_AREA_IDS]


###


def get_IC_def_targ_pop_req_mod(targ_pop_rec: dict, mod_mst_ID: str):
    return mod_mst_ID in targ_pop_rec[icmvf.IC_TARG_POP_DEF_REQ_MOD_MST_IDS]


def add_IC_def_targ_pop_req_mod(targ_pop_rec: dict, mod_mst_ID: str):
    targ_pop_rec[icmvf.IC_TARG_POP_DEF_REQ_MOD_MST_IDS].append(mod_mst_ID)
    return targ_pop_rec


def remove_IC_def_targ_pop_applic_req_mod(targ_pop_rec: dict, mod_mst_ID: str):
    targ_pop_rec[icmvf.IC_TARG_POP_DEF_REQ_MOD_MST_IDS].remove(mod_mst_ID)
    return targ_pop_rec


#######################################################################################################
#                                                                                                     #
#                                       Alternative PIN age group set 1                               #
#                                                                                                     #
#######################################################################################################


def get_IC_alt_PIN_age_group_set_1_rec(get_IC_alt_PIN_age_group_set_1_records: list, interv_ID: str):
    get_IC_alt_PIN_age_group_set_1_recs = [
        rec
        for rec in get_IC_alt_PIN_age_group_set_1_records
        if rec[icmvf.IC_ALT_PIN_AGE_GROUP_SET_1_INTERV_ID] == interv_ID
    ]
    return get_IC_alt_PIN_age_group_set_1_recs[0] if len(get_IC_alt_PIN_age_group_set_1_recs) > 0 else gbc.GB_NOT_FOUND

    # num_records = len(get_IC_alt_PIN_age_group_set_1_list)

    # stop = False
    # i = 0
    # while (i < num_records - 1) and (not stop):
    #     alt_PIN_age_group_set_1_rec = get_IC_alt_PIN_age_group_set_1_list[i]

    #     interv_ID_i = alt_PIN_age_group_set_1_rec[icmvf.IC_ALT_PIN_AGE_GROUP_SET_1_INTERV_ID]

    #     if interv_ID_i == interv_ID:
    #         stop = True
    #     else:
    #         i = i + 1

    # if stop:
    #     return alt_PIN_age_group_set_1_rec[i]
    # else:
    #     return gbc.GB_NOT_FOUND


def get_IC_alt_PIN_age_group_set_1(alt_PIN_age_group_set_1_rec: dict, t: int, a: int):
    return alt_PIN_age_group_set_1_rec[icmvf.IC_ALT_PIN_AGE_GROUP_SET_1_VALUE][a - 1][t]


#######################################################################################################
#                                                                                                     #
#                                       Alternative PIN age group set 2                               #
#                                                                                                     #
#######################################################################################################


def get_IC_alt_PIN_age_group_set_2_rec(get_IC_alt_PIN_age_group_set_2_list: list, interv_ID: str):
    get_IC_alt_PIN_age_group_set_2_recs = [
        rec
        for rec in get_IC_alt_PIN_age_group_set_2_list
        if rec[icmvf.IC_ALT_PIN_AGE_GROUP_SET_2_INTERV_ID] == interv_ID
    ]
    return get_IC_alt_PIN_age_group_set_2_recs[0] if len(get_IC_alt_PIN_age_group_set_2_recs) > 0 else gbc.GB_NOT_FOUND

    # num_records = len(get_IC_alt_PIN_age_group_set_2_list)

    # stop = False
    # i = 0
    # while (i < num_records - 1) and (not stop):
    #     alt_PIN_age_group_set_2_rec = get_IC_alt_PIN_age_group_set_2_list[i]

    #     interv_ID_i = alt_PIN_age_group_set_2_rec[icmvf.IC_ALT_PIN_AGE_GROUP_SET_2_INTERV_ID]

    #     if interv_ID_i == interv_ID:
    #         stop = True
    #     else:
    #         i = i + 1

    # if stop:
    #     return alt_PIN_age_group_set_2_rec[i]
    # else:
    #     return gbc.GB_NOT_FOUND


def get_IC_alt_PIN_age_group_set_2(alt_PIN_age_group_set_2_rec: dict, t: int, a: int):
    return alt_PIN_age_group_set_2_rec[icmvf.IC_ALT_PIN_AGE_GROUP_SET_2_VALUE][a - 1][t]


#######################################################################################################
#                                                                                                     #
#                                   Disease-specific costing  / TB costing                            #
#                                                                                                     #
#######################################################################################################


def get_IC_outpat_visit_inpat_day_unit_cost_DSC(
    outpat_visit_inpat_day_unit_cost_list, visit_type_curr_ID, deliv_chan_curr_ID, prog_area_curr_ID
):
    return outpat_visit_inpat_day_unit_cost_list[visit_type_curr_ID - 1, deliv_chan_curr_ID - 1, prog_area_curr_ID - 1]


def set_IC_outpat_visit_inpat_day_unit_cost_DSC(
    outpat_visit_inpat_day_unit_cost_list, visit_type_curr_ID, deliv_chan_curr_ID, prog_area_curr_ID, value
):
    outpat_visit_inpat_day_unit_cost_list[visit_type_curr_ID - 1, deliv_chan_curr_ID - 1, prog_area_curr_ID - 1] = value


#######################################################################################################
#                                                                                                     #
#                                       Comprehensive vs Specific Costing                             #
#                                                                                                     #
#######################################################################################################


def get_IC_outpat_visit_inpat_day_costs_CVS(outpat_visit_inpat_day_costs_list, visit_type_curr_ID, deliv_chan_curr_ID):
    return outpat_visit_inpat_day_costs_list[visit_type_curr_ID - 1, deliv_chan_curr_ID - 1]


def set_IC_outpat_visit_inpat_day_costs_CVS(
    outpat_visit_inpat_day_costs_list, visit_type_curr_ID, deliv_chan_curr_ID, value
):
    outpat_visit_inpat_day_costs_list[visit_type_curr_ID - 1, deliv_chan_curr_ID - 1] = value


#######################################################################################################
#                                                                                                     #
#                                       Results                                                       #
#                                                                                                     #
#######################################################################################################

###


def get_IC_num_services_by_deliv_chan(num_services_by_deliv_chan_array, interv_curr_ID, deliv_chan_curr_ID, t):
    return num_services_by_deliv_chan_array[interv_curr_ID - 1, deliv_chan_curr_ID - 1, t]


def set_IC_num_services_by_deliv_chan(num_services_by_deliv_chan_array, interv_curr_ID, deliv_chan_curr_ID, t, value):
    num_services_by_deliv_chan_array[interv_curr_ID - 1, deliv_chan_curr_ID - 1, t] = value



# Test 2

from datetime import date


import SpectrumCommon.Const.LG.LGConst as lgc
import SpectrumCommon.Const.LG.LGTags as lgmvt
import SpectrumCommon.Modvars.LG.LGModVarDefs as lgmvd
import SpectrumCommon.Modvars.LG.LGModVarUtil as lgmvu
import SpectrumCommon.Modvars.LG.LGReadDatabases as lgrdb
# import SpectrumCommon.Util.LG.LGTransactionCalcUtil as lgtcu


def init_modvars(countryID, first_year, final_year, extra):
    """
    Initializes Logistic MVs for new IHT projections.

        Parameters:
            countryID (string):  Ex: BEN for Benin.

            first_year (int): First year selected by user.

            final_year (int): Final year selected by user.

            extra (dict): Any other fields.

        Returns:
            mod_vars_rec (dict): MVs.
    """

    # Load databases
    # IHT_costing_on = extra[gbc.GB_IHT_ON] if gbc.GB_IHT_ON in extra else False
    # TB_costing_on = extra[gbc.GB_TB_COSTING_ON] if gbc.GB_TB_COSTING_ON in extra else False

    # facility_med_equip_group_DB = isrd.IS_load_facility_med_equip_group_DB()
    # facility_equip_input_DB = isrd.IS_load_facility_equip_input_DB()

    value_bool_false = False

    value_str = ""
    value_date = (date.today().strftime("%B %d, %Y"),)

    num_years_inputs = final_year - first_year + 1

    default_cold_chain_records_DB = lgrdb.LG_load_cold_chain_DB()

    supply_chain_level_records = lgmvd.LG_init_supply_chain_level_records(num_years_inputs)
    num_supply_chain_levels = lgmvu.get_LG_num_supply_chain_levels(supply_chain_level_records)
    vehicle_type_records = lgmvd.LG_init_vehicle_type_records(num_supply_chain_levels, num_years_inputs)
    worker_type_records = lgmvd.LG_init_worker_type_records(
        num_supply_chain_levels, vehicle_type_records, num_years_inputs
    )
    cold_chain_records = lgmvd.LG_init_cold_chain_records(num_supply_chain_levels, num_years_inputs)
    default_cold_chain_records = lgmvd.LG_init_default_cold_chain_records(default_cold_chain_records_DB)

    mod_vars_rec = {
        # Inputs / pre-result / calculated
        lgmvt.LG_VERSION_TAG: 1,
        lgmvt.LG_DATA_CHANGED_TAG: value_bool_false,
        lgmvt.LG_PROJ_VALID_TAG: value_bool_false,
        lgmvt.LG_FILE_NAME_TAG: value_str,
        lgmvt.LG_PROJ_VALID_DATE_TAG: value_date,
        lgmvt.LG_SUPPLY_CHAIN_LEVEL_RECORDS_TAG: supply_chain_level_records,
        lgmvt.LG_WAREHOUSE_RECORDS_TAG: [],
        lgmvt.LG_DISAGGREGATION_LEVEL_TAG: lgc.LG_NATIONAL_NO_DISAGGREGATION,
        lgmvt.LG_CALC_APPROACH_TAG: lgc.LG_CALC_APPROACH_COSTING_MST_ID,
        lgmvt.LG_VEHICLE_TYPE_RECORDS_TAG: vehicle_type_records,
        lgmvt.LG_WORKER_TYPE_RECORDS_TAG: worker_type_records,
        lgmvt.LG_COLD_CHAIN_RECORDS_TAG: cold_chain_records,
        lgmvt.LG_DEFAULT_COLD_CHAIN_RECORDS_MEGA_TAG: default_cold_chain_records,
        lgmvt.LG_THIRD_PARTY_LOG_CONTRACT_RECORDS_TAG: [],
        lgmvt.LG_ENTER_COMP_OP_COSTS_WAREHOUSES_TAG: value_bool_false,
        lgmvt.LG_ENTER_COMP_CONSTRUCT_COSTS_WAREHOUSES_TAG: value_bool_false,
        lgmvt.LG_ENTER_COMP_MAINT_OP_COSTS_VEHICLES_TAG: value_bool_false,
        lgmvt.LG_COST_FUEL_PER_LITER_TAG: 0.0,
        lgmvt.LG_VEHICLE_TARG_CONFIG_OPTION_TAG: lgc.LG_TARGET_CONFIG_WAREHOUSE_STD_MST_ID,
        lgmvt.LG_VEHICLE_TARG_VARY_TRIPS_PER_YEAR_TAG: value_bool_false,
        lgmvt.LG_COLD_CHAIN_TARG_CONFIG_OPTION_TAG: lgc.LG_TARGET_CONFIG_WAREHOUSE_STD_MST_ID,
        lgmvt.LG_VEHICLE_PERFORMANCE_CONFIG_OPTION_TAG: lgc.LG_PERFORMANCE_CONFIG_OPTION_ROUTE_BASED,
        lgmvt.LG_NUM_WORK_HOURS_PER_DAY: 8,
        lgmvt.LG_NUM_WORK_DAYS_PER_YEAR: 220,
    }

    # Debugging
    # with open('C:\Proj\IHTSampleFile.JSON', 'w') as f:
    #   json.dump(mod_vars_dict, f)

    return mod_vars_rec



import SpectrumCommon.Const.IC.ICTargetPopulationIDs as ictpi

# These delivery channels get tacked on at the end of the IH delivery channels and cannot be changed by the user.
IC_WASH_CHAN_CURR_ID = 1
IC_OTHER_NON_HEALTH_CHAN_CURR_ID = 2
IC_PRIVATE_SECTOR_CHAN_CURR_ID = 3
IC_NUM_DELIV_CHANS = 3

# Constants reflecting which year user wants to display *)
IC_CURRENT_YEAR = 1
IC_TARGET_YEAR = 2
IC_NUM_YEAR_ENDPOINTS = IC_TARGET_YEAR

IC_OUTPATIENT_VISIT_CURR_ID = 1
IC_INPATIENT_DAY_CURR_ID = 2
IC_NUM_VISIT_TYPES = IC_INPATIENT_DAY_CURR_ID
IC_ALL_VISIT_TYPES = IC_NUM_VISIT_TYPES + 1

IC_OUTPATIENT_VISIT_MST_ID = 1
IC_INPATIENT_DAY_MST_ID = 2

# Database latest versions
IC_DRUG_SUPPLY_DB_CURR_VERSION = "V11"
IC_TREATMENT_INPUT_DB_CURR_VERSION = "V12"
IC_PIN_DB_CURR_VERSION = "V7"
IC_PIN_BY_YEAR_DB_CURR_VERSION = "V1"
IC_NON_IMPACT_COVERAGE_DB_CURR_VERSION = "V2"
IC_NON_IMPACT_COVERAGE_BY_YEAR_DB_CURR_VERSION = "V1"
IC_TARG_POP_DB_CURR_VERSION = "V13"
IC_TARG_POP_DIRECT_ENTRY_DB_CURR_VERSION = "V1"
IC_ADOLESCENT_HEALTH_PIN_DB_CURR_VERSION = "V1"
IC_ADOLESCENT_HEALTH_COV_DB_CURR_VERSION = "V1"
IC_OTHER_COSTS_DB_CURR_VERSION = "V1"
# Databases for TB
IC_OUTPAT_VISIT_INPAT_DAY_COSTS_TB_DB_CURR_VERSION = "V1"
# Databases for LiST Costing
IC_OUTPAT_VISIT_INPAT_DAY_COSTS_CS_DB_CURR_VERSION = "V3"
IC_HEALTH_SYSTEM_COSTS_DB_CURR_VERSION = "V1"
# Databases for calculated things created in ICProj
IC_DRUG_COST_PER_AVG_CASE_DB_CURR_VERSION = "V1"
IC_SUPPLY_COST_PER_AVG_CASE_DB_CURR_VERSION = "V1"
IC_MED_PERSONNEL_MIN_PER_AVG_CASE_DB_CURR_VERSION = "V1"

# Database names (also the names of the sheets in ICModData.xlsx)
IC_DRUG_SUPPLY_DB_NAME = "DrugSupplyDB"
IC_TREATMENT_INPUT_DB_NAME = "TreatmentInputDB"
IC_TREATMENT_INPUT_GROUP_DB_NAME = "TreatmentInputGroupDB"  # Derived from TreatmentInputDB in ICModData but written to separate file for ease of processing.
IC_PIN_DB_NAME = "PIN_DB"
IC_PIN_BY_YEAR_DB_NAME = "PIN_ByYearDB"
IC_NON_IMPACT_COVERAGE_DB_NAME = "NonImpactCoverageDB"
IC_NON_IMPACT_COVERAGE_BY_YEAR_DB_NAME = "NonImpactCoverageByYearDB"
IC_TARG_POP_DB_NAME = "TargPopDB"
IC_TARG_POP_DIRECT_ENTRY_DB_NAME = "TargPopDirectEntryDB"
IC_ADOLESCENT_HEALTH_PIN_DB_NAME = "AdolescentHealthPIN_DB"
IC_ADOLESCENT_HEALTH_COV_DB_NAME = "AdolescentHealthCovDB"
# Databases for TB
IC_OUTPAT_VISIT_INPAT_DAY_COSTS_TB_DB_NAME = "OutpatVisitInpatDayCostsTBDB"
# Databases for LiST Costing
IC_OUTPAT_VISIT_INPAT_DAY_COSTS_CS_DB_NAME = "OutpatVisitInpatDayCostsCSDB"
IC_HEALTH_SYSTEM_COSTS_DB_NAME = "HealthSystemCostsDB"
# Databases for calculated things created in ICProj
IC_DRUG_COST_PER_AVG_CASE_DB_NAME = "DrugCostPerAvgCaseDB"
IC_SUPPLY_COST_PER_AVG_CASE_DB_NAME = "SupplyCostPerAvgCaseDB"
IC_MED_PERSONNEL_MIN_PER_AVG_CASE_DB_NAME = "MedPersonnelMinPerAvgCase"

# Database directories for data by country. Files in these folders are named using the 3 letter iso country codes instead of the database names above.
IC_NON_IMPACT_COVERAGE_BY_YEAR_DB_DIR = r"NonImpactCoverageByYear\countries"
IC_NON_IMPACT_COVERAGE_DB_DIR = r"NonImpactCoverage\countries"
IC_PIN_BY_YEAR_DB_DIR = r"PINsByYear\countries"
IC_PIN_DB_DIR = r"PINs\countries"
IC_TARG_POP_DIRECT_ENTRY_DB_DIR = r"TargPopDirectEntry\countries"
IC_OUTPAT_VISIT_INPAT_DAY_COSTS_TB_DB_DIR = r"OutpatVisitInpatDayCostsTBDB\countries"
IC_OUTPAT_VISIT_INPAT_DAY_COSTS_CS_DB_DIR = r"OutpatVisitInpatDayCostsCSDB\countries"
IC_HEALTH_SYSTEM_COSTS_DB_DIR = r"HealthSystemCostsDB\countries"

IC_DRUG_CURR_ID = 1
IC_SUPPLY_CURR_ID = 2
IC_NUM_DRUG_SUPPLY_TYPES = IC_SUPPLY_CURR_ID

IC_DRUG_MST_ID = 1
IC_SUPPLY_MST_ID = 2

# Treatment input item type master IDs
IC_TI_ALL_TREAT_INPUT_TYPES_MST_ID = 0
IC_TI_DRUGS_SUPPLIES_MST_ID = 1
IC_TI_MED_PERSONNEL_TIME_MST_ID = 2
IC_TI_OUTPAT_VISITS_MST_ID = 3  # defunct
IC_TI_INPAT_DAYS_MST_ID = 4  # defunct
IC_TI_SUPPLIES_MST_ID = 5  # defunct
IC_TI_PERSONNEL_TIME_MST_ID = 6  # defunct
IC_TI_EQUIPMENT_MST_ID = 7  # defunct
IC_TI_OTHER_MST_ID = 8  # defunct
IC_TI_OUTPAT_VISITS_INPAT_DAYS_MST_ID = 9

IC_TI_TREAT_INPUT_TYPES = [
    IC_TI_DRUGS_SUPPLIES_MST_ID,
    IC_TI_MED_PERSONNEL_TIME_MST_ID,
    IC_TI_OUTPAT_VISITS_INPAT_DAYS_MST_ID,
]

IC_TP_INTERV_MST_ID = 1
IC_TP_TARG_POP_MST_ID = 2

# Age group 1: <50, 50-59, 60-69, 70+
IC_AGE_GROUP_1_LT_50_CURR_ID = 1
IC_AGE_GROUP_1_50_TO_59_CURR_ID = 2
IC_AGE_GROUP_1_60_TO_69_CURR_ID = 3
IC_AGE_GROUP_1_70_PLUS_CURR_ID = 4
IC_NUM_AGE_GROUPS_SET_1 = IC_AGE_GROUP_1_70_PLUS_CURR_ID

# Age group 2: 0-4, 5-14, 15-29, 30-44, 45-59, 60-69, 70-79, 80+
IC_AGE_GROUP_2_0_TO_4_CURR_ID = 1
IC_AGE_GROUP_2_5_TO_14_CURR_ID = 2
IC_AGE_GROUP_2_15_TO_29_CURR_ID = 3
IC_AGE_GROUP_2_30_TO_44_CURR_ID = 4
IC_AGE_GROUP_2_45_TO_59_CURR_ID = 5
IC_AGE_GROUP_2_60_TO_69_CURR_ID = 6
IC_AGE_GROUP_2_70_TO_79_CURR_ID = 7
IC_AGE_GROUP_2_80_PLUS_CURR_ID = 8
IC_NUM_AGE_GROUPS_SET_2 = IC_AGE_GROUP_2_80_PLUS_CURR_ID

# Intervention special cases constants

IC_CS_IMPACT = "<CS Impact>"  # Coverage for intervention comes from CS
IC_AM_IMPACT = "<AM Impact>"  # Coverage for intervention comes from AM
IC_RN_IMPACT = "<RN Impact>"  # Coverage for intervention comes from RN
IC_FP_IMPACT = "<FP Impact>"  # Coverage for intervention comes from FP
IC_TB_IMPACT = "<TB Impact>"  # Coverage for intervention comes from TB
IC_MA_IMPACT = "<MA Impact>"  # Coverage for intervention comes from MA
IC_NC_IMPACT = "<NC Impact>"  # Coverage for intervention comes from NC
IC_IS_PROG_MGMT = "<IS Programme Management>"
IC_RN_PC_HIV = "<RN PC HIV>"
IC_RN_STIS = "<RN STIs>"
IC_NOT_DELIV_MAIN_CHANNELS = (
    "<Not Delivered in Main Delivery Channels>"  # Not delivered in IH delivery channels, only IC delivery channels.
)
IC_BREAST_CANCER_TREATMENT = "<Breast Cancer Treatment>"
IC_TB_NOTIF_TREATMENT = "<TB Notification and Treatment>"
IC_CS_CHILDBIRTH = "<CS Childbirth>"
IC_PREG_WOMEN_MALARIA = "<Pregnant Women Malaria>"
IC_ITN_IRS_MALARIA = "<ITM and IRS Malaria>"
IC_MALARIA_PSEUDO_IMPACT = "<Malaria Pseudo Impact>"
IC_ART_SET_1 = "<ART Set 1>"
IC_ART_SET_2 = "<ART Set 2>"
IC_PC_MNS = "<NC PC MNS>"
IC_NC_CANCER = "<NC Cancer>"
IC_BASIC_INTENS_DEPRESSION = "<Basic Intensive Depression>"
IC_BASIC_INTENS_PSYCHOSIS = "<Basic Intensive Psychosis>"
IC_BASIC_INTENS_BIPOLAR = "<Basic Intensive Bipolar>"
IC_BASIC_INTENS_DEV_DISORDERS = "<Basic Intensive Developmental Disorders>"
IC_BASIC_INTENS_BEHAV_DISORDERS = "<Basic Intensive Behavioral Disorders>"
IC_PENTA_VACC = "<Pentavalent Vaccine>"
IC_CS_STANDARD_VACC = "<CS Standard Vaccines>"
IC_CS_DET_VACC = "<CS Detailed Vaccines>"
IC_DEV_MODE_VACC = "<CS Developer Mode Vaccines>"
IC_BREAST_CANCER_TR_PINS = "<Breast Cancer Treatment PINs>"
IC_WORKERS_HEALTH = "<Workers Health>"  # Intervention should only show up for Worker's Health mode.
IC_ADOLESCENT_HEALTH = "<Adolescent Health>"  # Intervention should only show up for Adolescent Health mode.
IC_TB_DAT_DISAG = "<TB DAT Disaggregate>"
IC_TB_DAT_AG = "<TB DAT Aggregate>"
IC_PHC = "<PHC>"  # Intervention should only show up for PHC mode.
IC_ALT_COVERAGE = "<Alternative Coverage>"
IC_HIDE_CONFIG = (
    "<Hide Configuration>"  # Hide intervention in configuration (where interventions are selected, added, etc.)
)
IC_HIDE_INPUT_OUTPUT_DATA = "<Hide Input Output Data>"  # Essentially show intervention as label everywhere.
IC_HIDE_INPUT_DATA = "<Hide Input Data>"  # Show intervention as label in inputs, with data next to it in results.
IC_HIDE_PINS = "<Hide PINs>"  # Hide intervention in PINs table.
IC_HIDE_TARG_POPS = "<Hide Target Populations>"  # Hide intervention in Target population table.
IC_HIDE_COVERAGES = "<Hide Coverages>"  # Hide intervention in Coverages table.
IC_LOCK_COVERAGES = "<Lock Coverages>"  # Lock intervention in Coverages table.
IC_HIDE_TREAT_INPUTS = "<Hide Treatment Inputs>"  # Hide intervention in Treatment inputs table.
IC_HIDE_DELIV_CHANNELS = "<Hide Delivery Channels>"  # Hide intervention in Delivery channels table.
IC_HIDE_RESULTS = "<Hide Results>"  # Hide intervention in IC results tables.
IC_CALC_PINS_ALL_YEARS_LOCKED = "<Calculate PINs for All Years, Locked>"  # PINs are calculated at all years but locked in table (cannot be changed).
IC_CALC_PINS_ALL_YEARS_OVERRIDE = "<Calculate PINs for All Years, Overridable>"  # PINs are calculated at all years, but user can override them in table.
IC_CALC_PINS_TWO_PLUS_LOCKED = (
    "<Calculate PINs for Year Two Forward>"  # PINs are calculated and locked from the second year forward.
)
IC_MULTIPLY_PINS_100 = "<Multiply PINs by 100>"  # PINs need to be multiplied by 100 in table.
IC_PINS_OVER_100 = "<PINs Over 100>"  # PINs can be > 100 in table.
IC_MALARIA_PINS = "<Malaria PINs>"

# Maximum number of custom interventions that the user can have.
IC_MaxCustomInterventions = 500

# Minimum master ID that a UHM custom intervention must have in order to
# avoid conflicting with the master IDs in MstIVDB.csv in IV and the
# minimum master ID that a LiST custom intervention must have.
IC_MinCustIntervMstID = 10000

# Target populations that use the number of services of various interventions in their calculation
IC_SERVICE_TARG_POPS = [
    ictpi.IC_TP_PAT_6_MO_ISON_RESIST_SHORTER_SEC_LINE_15_PLUS_MST_ID,
    ictpi.IC_TP_PAT_6_MO_ISON_RESIST_SHORTER_SEC_LINE_0_TO_14_MST_ID,
    ictpi.IC_TP_PAT_LONGER_REG_MDR_RR_TB_XDR_MST_ID,
    ictpi.IC_TP_PAT_6_MO_ISON_RESIST_SHORTER_SEC_LINE_MST_ID,
]

CPR_GOAL_CURR_ID = 1
UNMET_NEED_GOAL_CURR_ID = 2
NUM_FAM_PLAN_GOALS = 2

CPR_GOAL_MST_ID = 1
UNMET_NEED_GOAL_MST_ID = 2



from SpectrumCommon.Const.IV.IVConst import *
from SpectrumCommon.Const.IV.IVDatabaseKeys import *

from SpectrumCommon.Const.IH.IHConst import *
from SpectrumCommon.Const.IH.IHTags import *

from SpectrumCommon.Const.CS.CSTags import *

#############################    IV    ##########################

PC_IV_GROUP_DB_CURR_VERSION = IV_IH_GROUP_DB_CURR_VERSION

PC_IV_GROUP_DB_NAME         = IV_GROUP_DB_NAME

# Group DB (GDB)

PC_IV_GROUP_MST_ID_KEY_GDB     = IV_GROUP_MST_ID_KEY_GDB
PC_IV_SUBGROUP_MST_ID_KEY_GDB  = IV_SUBGROUP_MST_ID_KEY_GDB   

#############################    IH    ##########################

PC_IH_PROG_AREA_MST_ID                  = IH_PROG_AREA_MST_ID
PC_IH_MODULE_MST_ID                     = IH_MODULE_MST_ID  
PC_IH_HEALTH_SYS_MST_ID                 = IH_HEALTH_SYS_MST_ID
PC_IH_HEALTH_INFORM_SYSTEMS_HSPM_MST_ID = IH_HEALTH_INFORM_SYSTEMS_HSPM_MST_ID
PC_IH_HEALTH_WORKFORCE_HSPM_MST_ID      = IH_HEALTH_WORKFORCE_HSPM_MST_ID
PC_IH_INFRASTRUCTURE_HSPM_MST_ID        = IH_INFRASTRUCTURE_HSPM_MST_ID
PC_IH_LOGISTICS_HSPM_MST_ID             = IH_LOGISTICS_HSPM_MST_ID
PC_IH_PPLI_MST_ID                       = IH_PPLI_MST_ID
PC_IH_SOCIAL_ENABLERS_MST_ID            = IH_SOCIAL_ENABLERS_MST_ID
PC_IH_NATIONAL_HSAU_MST_ID              = IH_NATIONAL_HSAU_MST_ID 
PC_IH_PROVINCE_HSAU_MST_ID              = IH_PROVINCE_HSAU_MST_ID
PC_IH_DISTRICT_HSAU_MST_ID              = IH_DISTRICT_HSAU_MST_ID

PC_IH_HIV_AIDS_MST_ID = IH_HIV_AIDS_MST_ID
PC_IH_NCD_MST_ID      = IH_NCD_MST_ID

PC_IH_PROG_AREA_RECORDS_TAG_V1 = IH_PROG_AREA_RECORDS_TAG_V1
PC_IH_HEALTH_SYS_ADMIN_UNIT_RECORDS_TAG = IH_HEALTH_SYS_ADMIN_UNIT_RECORDS_TAG

#############################    CS    ##########################

PC_CS_ANNUAL_PERC_GROWTH_RATE_TAG = CS_AnnualPercGrowthRateTag




import math
import numpy as np
from scipy import interpolate
from .make_model2 import make_model2
from .goveqs_basis2 import goveqs_basis2
from .goveqs_scaleup import goveqs_scaleup
from .ode15s import ode15s
from .goveqs_basissmall import goveqs_basissmall
from .make_smallmodel import make_smallmodel
import collections.abc

#def get_objective(x, prm, ref, sel, agg, gps, gps1, gps2, calfn):
def get_objective(prm, ref, sel, agg, gps, gps1, gps2, calfn): # Signature changed on June 09, 2024, after Carel suggested that we stop passing xs
    
    r = prm['r']
    p = prm['p']
    i = ref['i']
    s = ref['s']
    xi = ref['xi']
  
    #%i0 = ref.i0; s0 = ref.s0; 
    i1 = ref['i1']; s1 = ref['s1']; i2 = ref['i2']; s2 = ref['s2']


    r['cs_1997'] = np.array(r['cs_1997'])
    r['cs_2022'] = np.array(r['cs_2022'])
    r['cs2_2022'] = np.array(r['cs2_2022'])
    r['cs2_1997'] = np.array(r['cs2_1997'])


    # First, check whether the proposed parameter ranges lie within the
    # specified ranges (i.e. implement the uniform priors - can also be done
    # after performing the simulations and alongside the likelihood, but this
    # way saves time, rejecting out-of-range parameters without needing to do
    # the simulations

    #mat = np.array([prm['bounds'][:len(x),1]-x, x-prm['bounds'][:len(x),0]])
    #mat = np.array([prm['bounds'][1,:]-x, x-prm['bounds'][0,:]])
    #The following line was commented out on June 09, 2024 because Carel suggested that we stop passing x to get_objective
    #mat = np.array([np.array(prm['bounds'])[1,range(len(x))]-x, x-np.array(prm['bounds'])[0,range(len(x))]]) #mat = [prm.bounds(2,1:length(x))-x; x-prm.bounds(1,1:length(x))];

    # All values should be positive if x is within the specified ranges

#if np.amin(mat)< 0:
#    
#    out = math.inf*calfn['sgn']
#    # +Inf if using sum of squares, -Inf if using likelihoods
#    aux = None
##     keyboard;
#else:
        
    # Note the use of the subfunction alloc_parameters, to keep things tidy
    # with the allocation of parameters
    #(r, p) = alloc_parameters(x,r,p,xi)
        
    # --- Set up the necessary models -------------------------------------
    #% Equilibrium model - no MDR, no HIV
    p0 = p.copy()
    r0 = r.copy()
        #p0['pu'] = 0
        #p0['rfbeta_mdr'] = 0 ;#p0['rfbeta_hiv'] = 0; #r0['MDR_acqu'] = 0
        #r0['ART_init'] = 0 #p0['MDR_rec'] = np.array([0,0])
        #
    p0['pu'] = 0 
    r0['cs']  = r['cs_1997'] #r0.cs  = r.cs_1997; 
    r0['cs2'] = r['cs2_1997'] #r0.cs2 = r.cs2_1997;
    p0['rfbeta_mdr'] = 0 #p0.rfbeta_mdr = 0; 
    r0['MDR_acqu'] = 0 #r0.MDR_acqu = 0;  
    p0['MDR_rec'] = np.array([0, 0]) #p0.MDR_rec = [0 0]; 
    p0['rfbeta_hiv'] = 0 #p0.rfbeta_hiv = 0; 
    r0['ART_init'] = 0 #r0.ART_init = 0; 
        #
        #Check where make_model2 is defined
    M0 = make_smallmodel(p0, r0, i1, s1, gps1,prm) #M0 = make_model(p0, r0, i, s, gps) #make_model2(p0, r0, i, s, gps)

        #% Introduction of MDR, no ART (1970-1997),
    p1 = p.copy(); r1 = r.copy()#p1 = p; r1 = r;
        #p1['pu'] = 0; r1['ART_init'] = 0; p1['MDR_rec'] = np.array([0, 0]) #p['MDR_rec2015'] #p1.pu = 0; r1.ART_init = 0; p1.MDR_rec = p.MDR_rec2015;
        
    p1['pu'] = 0; r1['cs']  = r['cs_1997']; r1['cs2'] = r['cs2_1997'] #p1.pu = 0; r1.cs  = r.cs_1997; r1.cs2 = r.cs2_1997;
    p1['MDR_rec'] = np.array([0, 0]) #p1.MDR_rec = [0 0];
    p1['rfbeta_hiv'] = 0; r1['ART_init'] = 0#p1.rfbeta_hiv = 0; r1.ART_init = 0;    % I guess HIV should introduce here i.e p1.rfbeta_hiv

    M1a = make_smallmodel(p1, r1, i2, s2, gps2,prm)   
    M1 = make_model2(p1, r1, i, s, gps,prm) #; % To transfer it into full matrix

        #% Start of RNTCP to start of ART  (prm.ART_start(here 2004 for India) (RNTCP scale up 1997 - 2007)
    p2 = p.copy(); r2 = r.copy()
        #r2['ART_init'] = 0; p2['MDR_rec'] = np.array([0, 0]); #%p.MDR_rec2015;  
    p2['pu'] = 0 + (p['pu_2022'] - 0)*((prm['ART_start']-1997)/25)#p2.pu = 0 + (p.pu_2022 - 0)*((prm.ART_start-1997)/25);  % Assuming linear scaleup from 1997 to 2022
    r2['cs']  = r['cs_1997'] + (r['cs_2022'] - r['cs_1997'])*((prm['ART_start']-1997)/25)#r2.cs  = r.cs_1997 + (r.cs_2022 - r.cs_1997)*((prm.ART_start-1997)/25);
    r2['cs2'] = r['cs2_1997'] + (r['cs2_2022'] - r['cs2_1997'])*((prm['ART_start']-1997)/25) #r2.cs2 = r.cs2_1997 + (r.cs2_2022 - r.cs2_1997)*((prm.ART_start-1997)/25); % Assumed to be same from 1997 to 2022
    p2['MDR_rec'] = np.array([0, 0]); r2['ART_init'] = 0 #p2.MDR_rec = [0 0]; r2.ART_init = 0; %p.MDR_rec2015;       
    M2 = make_model2(p2, r2, i, s, gps,prm)
    
        #% Start of ART to start of DST (scaleup 2003 - 2022)
    p3 = p.copy(); r3 = r.copy()
    p3['MDR_rec'] = p['MDR_rec2015']  # %p.MDR_rec2015    
 
    #Guy aligned this to Matlab on 2024-07-25 
    p3['pu'] = 0 + (p['pu_2022'] - 0)*((2015-1997)/25) #p3.pu = 0 + (p.pu_2022 - 0)*((2015-1997)/25);  % Assuming linear scaleup from 1997 to 2022
    r3['cs']  = r['cs_1997'] + (r['cs_2022'] - r['cs_1997'])*((2015-1997)/25)#r3.cs  = r.cs_1997 + (r.cs_2022 - r.cs_1997)*((2015-1997)/25);
    r3['cs2'] = r['cs2_1997'] + (r['cs2_2022'] - r['cs2_1997'])*((2015-1997)/25)#r3.cs2 = r.cs2_1997 + (r.cs2_2022 - r.cs2_1997)*((2015-1997)/25); % Assumed to be same from 1997 to 2022   
    #End #Guy aligned this to Matlab on 2024-07-25
    M3 = make_model2(p3, r3, i, s, gps,prm) #;
      
        #% Scale-up of DST (2015-2022)
    p4 = p.copy(); r4 = r.copy()
    p4['pu'] = p['pu_2022']; r4['cs']  = r['cs_2022']; r4['cs2'] = r['cs2_2022'] #p4.pu = p.pu_2022; r4.cs  = r.cs_2022; r4.cs2 = r.cs2_2022; 
    p4['MDR_rec'] = p['MDR_rec2022'] #p4.MDR_rec = p.MDR_rec2022;#Guy aligned this to Matlab on 2024-07-25
    M4 = make_model2(p4, r4, i, s, gps,prm)
        #%tic
        #% --- Simulate the models

        #% --- Simulate the models
        #% Equilibrium model
        #init = np.zeros((1,i['nx'])); seed = 1e-6 #init = zeros(1,i.nx); seed = 1e-6;
        #init[0,i['U']['h0']['v0']] = (1-seed)
        #init[0,i['I']['h0']['ds']] = seed#init(i.U.h0.v0) = (1-seed); init(i.I.h0.ds) = seed;
        ##geq = @(t,in)goveqs_basis2(t, in, M0, i, prm, sel, agg);
        ##[t0, soln0] = ode15s(geq, [0:2e3], init, odeset('NonNegative',[1:i.nstates]));
        #y0 = list(init[0])
        #(t0, soln0) = ode15s(goveqs_basis2, y0, np.linspace(0, 2e3,100), args=(M0, i, prm, sel, agg))

        #% Equilibrium model from the simplified small model
    init1 = np.zeros((1,i1['nx'])); seed = 1e-6
    init1[0][i1['U']['a1']['h0']['v0']] = prm['N'][0]/np.sum(prm['N']) 
    init1[0][i1['U']['a2']['h0']['v0']] = prm['N'][1]/sum(prm['N']) #init(i1.U.a2.h0.v0) = prm.N(2)/sum(prm.N); 
    init1[0][i1['U']['a3']['h0']['v0']] = prm['N'][2]/sum(prm['N']) #init(i1.U.a3.h0.v0) = prm.N(3)/sum(prm.N); 
    init1[0][i1['U']['a4']['h0']['v0']] = (prm['N'][3]/sum(prm['N']))-seed#init(i1.U.a4.h0.v0) = (prm.N(4)/sum(prm.N))-seed; 
    init1[0][i1['U']['a5']['h0']['v0']] = prm['N'][4]/sum(prm['N'])#init(i1.U.a5.h0.v0) = prm.N(5)/sum(prm.N); 
    init1[0][i1['I']['a4']['h0']['ds']] = seed#init(i1.I.a4.h0.ds) = seed;

    prm['growth'] = 0 #prm.growth = 0;
        #geq = @(t,in) goveqs_basissmall(t, in, r, M0, i1, prm);
        #[t0, soln0] = ode15s(geq, [0:2e3], init, odeset('NonNegative',[1:i1.nstates]));
    y0 = list(init1[0])
    (t0, soln0) = ode15s(goveqs_basissmall, y0, np.linspace(0, 2e3,2000), args=(r,M0, i1, prm))

    #Guy is updating this on 2024-07-25 to align with the code in NGA_final_vaccine
    #inds1 = np.intersect1d(np.intersect1d(s['U'], s['h0']),s['v0']) #inds1 = intersect(intersect(s.U, s.h0),s.v0);
    #inds2 = np.intersect1d(np.intersect1d(np.intersect1d([s['Lf'], s['Ls']], s['h0']), s['ds']),s['v0'])#inds2 = intersect(intersect(intersect([s.Lf, s.Ls], s.h0), s.ds),s.v0);
    #inds3 = np.intersect1d(np.intersect1d(np.concatenate([s['Isc'], s['I'], s['E'], s['Rlo'], s['Rhi'], s['R']]), s['h0']), s['ds']) #inds3 = intersect(intersect([s.Isc, s.I, s.E, s.Rlo, s.Rhi, s.R], s.h0), s.ds);
    #inds4 = np.intersect1d(np.intersect1d(np.intersect1d(np.concatenate([s['Dx'],s['Tx'],s['Tx2']]),s['ds']),s['h0']),s['pr']) #inds4 = intersect(intersect(intersect([s.Dx,s.Tx,s.Tx2],s.ds),s.h0),s.pr);
    #inds0  = np.concatenate([inds1, inds2, inds3,inds4]) #inds0  = [inds1, inds2, inds3,inds4];
    inds1 = np.intersect1d(np.intersect1d(s2['U'], s2['h0']),s2['v0']) #inds1 = intersect(intersect(s.U, s.h0),s.v0);
    inds2 = np.intersect1d(np.intersect1d(np.intersect1d([s2['Lf'], s2['Ls']], s2['h0']), s2['ds']),s2['v0'])#inds2 = intersect(intersect(intersect([s.Lf, s.Ls], s.h0), s.ds),s.v0);
    inds3 = np.intersect1d(np.intersect1d(np.concatenate([s2['Isc'], s2['I'], s2['E'], s2['Rlo'], s2['Rhi'], s2['R']]), s2['h0']), s2['ds']) #inds3 = intersect(intersect([s.Isc, s.I, s.E, s.Rlo, s.Rhi, s.R], s.h0), s.ds);
    inds4 = np.intersect1d(np.intersect1d(np.intersect1d(np.concatenate([s2['Dx'],s2['Tx'],s2['Tx2']]),s2['ds']),s2['h0']),s2['pr']) #inds4 = intersect(intersect(intersect([s.Dx,s.Tx,s.Tx2],s.ds),s.h0),s.pr);
    inds0  = np.concatenate([inds1, inds2, inds3,inds4]) #inds0  = [inds1, inds2, inds3,inds4];


    #% To allocate results from a small model to input vector for a bigger model
    #% model (60 to 115 compartments)
    init_smallmodel =  soln0[-1,:]
    init2 = np.zeros((1,i2['nx'])) #init = np.zeros((1,i['nx'])) #init = zeros(1,i.nx);
    init2[0][inds0] = init_smallmodel #init(inds0) = init_smallmodel;

    
    # % Introduction of MDR  
    prm['growth'] = prm['gth'][0] #prm.growth = prm.gth(1);
    #geq = @(t,in)goveqs_basissmall(t, in, r, M1a, i2, prm) #geq = @(t,in)goveqs_basissmall(t, in, r, M1a, i2, prm);  
    #[t1, soln1] = ode15s(geq, [1970 1997], init2, odeset('NonNegative',[1:i2.nstates])) #[t1, soln1] = ode15s(geq, [1970 1997], init2, odeset('NonNegative',[1:i2.nstates]));
    (t1, soln1) = ode15s(goveqs_basissmall, init2[0], np.linspace(1970, 1997), args=(r,M1a, i2, prm))

    ##End Guy is updating this on 2024-07-25 to align with the code in NGA_final_vaccine
   
    #% To initialise with MDR components
    inds1 = np.intersect1d(np.intersect1d(s['U'], s['h0']),s['v0']) #inds1 = intersect(intersect(s.U, s.h0),s.v0);
    inds2 = np.intersect1d(np.intersect1d([s['Lf'], s['Ls']], s['h0']),s['v0']) #inds2 = intersect(intersect([s.Lf, s.Ls], s.h0),s.v0);
    inds3 = np.intersect1d([s['Isc'], s['I'], s['E'], s['Rlo'], s['Rhi'], s['R']], s['h0'])
    inds4 = np.intersect1d(np.intersect1d([s['Dx'],s['Tx'],s['Tx2']],s['h0']),s['pr'])
    inds_full  = np.concatenate([inds1, inds2, inds3,inds4]) #inds_mdr  = [inds1, inds2, inds3,inds4];
    
    #init1 = init[0][inds_mdr]#init1 = init(inds_mdr);
  
      
    #prm['growth'] = prm['gth'][0] #prm.growth = prm.gth(1);
    #    #geq = @(t,in)goveqs_basissmall(t, in, r, M1a, i2, prm);  
    #    #[t1, soln1] = ode15s(geq, [1970 1997], init1, odeset('NonNegative',[1:i2.nstates]));
    #(t1, soln1) = ode15s(goveqs_basissmall, init1, np.linspace(1970, 1997), args=(r,M1a, i2, prm))

   
        #%     mat = zeros(i.nx,2);  % (Column 1 is i, column 2 is i1)
        #%     mat(:,1) = 1:i.nx;
        #%     mat(inds_mdr,2) = 1:length(inds_mdr);

        #% To allocate results from a small model to input vector for a bigger model
    init_smallmodel2 =  soln1[-1,:] #init_smallmodel2 =  soln1(end,:);
    init = np.zeros((1,i['nx'])) #init = zeros(1,i.nx);
    init[0][inds_full] = init_smallmodel2 #init(inds_mdr) = init_smallmodel2;
    
    #% Introduction of RNTCP % 
    #%init = init_smallmodel2; %soln1(end,:);
    #prm['growth'] = prm['gth'][1] #prm.growth = prm.gth(2);
        #geq = @(t,in) goveqs_scaleup(t, in, r, M1, M2, [1997 prm.ART_start], i, prm, sel, agg);#geq = @(t,in) goveqs_scaleup(t, in, r, M1, M2, [1997 2007], i, prm, sel, agg);
        #[t2, soln2] = ode15s(geq, [1997:prm.ART_start], init, odeset('NonNegative',[1:i.nstates]));
        #(t2, soln2) = ode15s(goveqs_scaleup, init[0], np.linspace(1997, prm['ART_start']), args=(r,M1, M2, [1997, 2007], i, prm, sel, agg))
    #(t2, soln2) = ode15s(goveqs_scaleup, init[0], np.linspace(1997, prm['ART_start']), args=(r,M1, M2, [1997, prm['ART_start']], i, prm, sel, agg))
    #% Before Introduction of RNTCP %
    prm['growth'] = 0 #prm.growth = 0; %prm.gth(2);  %2
    # %geq = @(t,in) goveqs_scaleup(t, in, r, M1, M2, [1997 prm.ART_start], i, prm, sel, agg);
    #geq = @(t,in) goveqs_basis2(t, in, r, M1, i, prm, sel, agg); %1e3
    #[t2a, soln2a] = ode15s(geq, [0:600], init, odeset('NonNegative',[1:i.nstates]));
    (t2a, soln2a) = ode15s(goveqs_basis2, init[0], np.linspace(0, 600), args=(r,M1, i, prm, sel, agg))
    #% Introduction of RNTCP %
    prm['growth'] = prm['gth'][1] #prm.growth = prm.gth(2);  %2
    #geq = @(t,in) goveqs_scaleup(t, in, r, M1, M2, [1997 prm.ART_start], i, prm, sel, agg);
    #%geq = @(t,in) goveqs_basis2(t, in, r, M1, i, prm, sel, agg);
    #[t2, soln2] = ode15s(geq, [1997:1:prm.ART_start], soln2a(end,:), odeset('NonNegative',[1:i.nstates]));
    (t2, soln2) = ode15s(goveqs_scaleup, soln2a[-1,:], np.linspace(1997, prm['ART_start']), args=(r,M1, M2, [1997, prm['ART_start']], i, prm, sel, agg))
    
        #% Introduction of ART (prm.ART_start(here 2004) onwards 2003-2015)
    init = soln2[-1,:] 
    prm['growth'] = prm['gth'][2] #prm.growth = prm.gth(3);
        #geq = @(t,in) goveqs_scaleup(t, in, r, M2, M3, [prm.ART_start 2022], i, prm, sel, agg);
        #geq = @(t,in) goveqs_scaleup(t, in, r, M2, M3, [prm.ART_start 2015], i, prm, sel, agg);#[t3, soln3] = ode15s(geq, [prm.ART_start:2015], init, odeset('NonNegative',[1:i.nstates]));
        #(t3, soln3) = ode15s(goveqs_scaleup, init, np.linspace(prm['ART_start'],2015), args=(r,M2, M3, [prm['ART_start'], 2022], i, prm, sel, agg))
    (t3, soln3) = ode15s(goveqs_scaleup, init, np.linspace(prm['ART_start'],2015), args=(r,M2, M3, [prm['ART_start'], 2015], i, prm, sel, agg))

        #% DST scale-up (Assuming from 2015 to 2022)
    init = soln3[-1,:]; prm['growth'] = prm['gth'][3] #init = soln3(end,:); prm.growth = prm.gth(4);
        #geq = @(t,in) goveqs_scaleup(t, in, r, M3, M4, [2015 2022], i, prm, sel, agg);#[t4, soln4] = ode15s(@(t,in) goveqs_scaleup(t, in,r, M3, M4, [2015 2022], i, prm, sel, agg), [2015 2023], init, odeset('NonNegative',[1:i.nstates]));
    (t4, soln4) = ode15s(goveqs_scaleup, init, np.linspace(2015,2023), args=(r,M3, M4, [2015, 2022], i, prm, sel, agg))

        #%keyboard;
    allt   = np.concatenate([t2, t3[1:], t4[1:]]) #allt   = [t2; t3(2:end); t4(2:end)]; #% ;   % 1997:2023
    allsol = np.concatenate([soln2, soln3[1:,:],soln4[1:,:]]) #allsol = [soln2; soln3(2:end,:);soln4(2:end,:)];
        
    proj_t = np.array(range(1997,2024))
    soln = np.zeros((len(proj_t),i['nx']))
    for jj in range(i['nx']):
        temp_fn_sol = interpolate.interp1d(allt, allsol[:,jj],fill_value="extrapolate") 
        soln[:,jj] = [temp_fn_sol(t) for t in proj_t]

        #soln = interp1(allt, allsol, 1997:2023) #soln = interp1(allt, allsol, 1997:2023);
        #dsol   = diff(soln,[],1);
        #sfin  = soln(end,:);
    dsol = np.diff(soln, 1, axis=0)
    sfin = soln[-1,]

        #%  toc
        #%  keyboard;
    #This line was added on 2024-10-23
    pop = np.sum(soln[0:(np.shape(soln)[0]-1),0:(i['nstates']-1)],1) # %1997:2022 (1:end -> 1997:2023) #pop = sum(soln(1:end-1,1:i.nstates),2);  %1997:2022 (1:end -> 1997:2023)

        #% --- Get the objectives ----------------------------------------------
    inc2000      = dsol[len(pop)-23,i['aux']['inc'][0]]/pop[len(pop)-23]*1e5 #  inc2000      = sum(dsol(end-22,i.aux.inc(1)))*1e5  
    if isinstance(inc2000, (collections.abc.Sequence, np.ndarray)):
        inc2000 = sum(inc2000) 

    inc2022  = dsol[-1,i['aux']['inc'][0]]/pop[-1]*1e5 #inc2022  = sum(dsol(end,i.aux.inc(1)))*1e5;
    if isinstance(inc2022, (collections.abc.Sequence, np.ndarray)):
        inc2022 = sum(inc2022)
    inc_h1   = dsol[-1,i['aux']['inc'][1]]/pop[-1]*1e5 #inc_h1   = dsol(end,i.aux.inc(2))*1e5;     %([2,3])
    mdr2015  = dsol[-8,i['aux']['inc'][2]]/pop[-8]*1e5 #mdr2015  = dsol(end-7,i.aux.inc(3))*1e5; %end--> 2022
    mdr2022  = dsol[-1,i['aux']['inc'][2]]/pop[-1]*1e5 #mdr2022  = dsol(end,i.aux.inc(3))*1e5;

    noti_cu      =  sum(np.sum(dsol[(len(proj_t)-23):len(proj_t),i['aux']['noti']]/pop[(len(proj_t)-23):len(proj_t)][:,None]*1e5,1)) #noti_cu      =  sum(sum(dsol(end-22:end,i.aux.noti)*1e5,2));  %cumulative notification from 2000 to 2022
    #noti_cu      =  sum(sum((dsol(end-22:end,i.aux.noti)./pop(end-22:end))*1e5,2));

    noti2022     = sum(dsol[-1,i['aux']['noti']]/pop[-1])*1e5 #noti     = sum(dsol(end,i.aux.noti))*1e5;
    ART_covg = sum(sfin[s['hart']]/np.sum(sfin[[s['h1'],s['hart']]])) #ART_covg = sum(sfin(s.hart)/sum(sfin([s.h1,s.hart])));
    HIV_prev = np.sum(sfin[[s['h1'], s['hart']]])/np.sum(sfin[0:(i['nstates']-1)]) #HIV_prev = sum(sfin([s.h1, s.hart]))/sum(sfin(1:i.nstates));
        
    mort_H02000  = dsol[len(proj_t)-23,i['aux']['mort'][0]]/pop[-23]*1e5 #mort_H02000  = dsol(end-22,i.aux.mort(1))*1e5;

    mort_H02022  = dsol[-1,i['aux']['mort'][0]]/pop[-1]*1e5 #mort_H0  = dsol(end,i.aux.mort(1))*1e5;
    mort_H12022  = dsol[-1,i['aux']['mort'][1]]/pop[-1]*1e5 #mort_H1  = dsol(end,i.aux.mort(2))*1e5;
    mdriniTX = dsol[-1,i['aux']['noti2']]/pop[-1]*1e5 #mdriniTX = dsol(end,i.aux.noti2)*1e5;  % MDR treatment initiation 2019
    sym_sim =  sum(sfin[s['I']])/sum(sfin[s['prevalent']]) #sym_sim =  sum(sfin(s.I))/sum(sfin(s.prevalent)); 
    inc2015  = dsol[-8,i['aux']['inc'][0]]/pop[-8]*1e5 #inc2015  = sum(dsol(end-7,i.aux.inc(1)))*1e5;
    if isinstance(inc2015, (collections.abc.Sequence, np.ndarray)):
        inc2015 = sum(inc2015)
    
    aux = {}
    if inc2022 < 1 : #if inc2022 < 1
        out = math.inf*calfn['sgn']
        aux = {}
    else:
            #out = calfn['fn'](inc2015, inc2022, inc_h1, noti, ART_covg, HIV_prev, mort_H0, mort_H1, mdr2015, mdr2022, mdriniTX, sym_sim) #out = calfn.fn(inc2015, inc2022, inc_h1, noti, ART_covg, HIV_prev, mort_H0, mort_H1, mdr2015, mdr2022, mdriniTX, sym_sim);
            #out = calfn.fn(sym_sim,inc2000, inc2022, mort_H02000,mort_H02022,noti_cu, noti2022, mdr2015, mdr2022, mdriniTX,HIV_prev,inc_h1,ART_covg,mort_H12022)
        out = calfn['fn'](sym_sim,inc2000, inc2022, mort_H02000,mort_H02022,noti_cu, noti2022, mdr2015, mdr2022, mdriniTX,HIV_prev,inc_h1,ART_covg,mort_H12022) #out = calfn.fn(inc2015, inc2022, inc_h1, noti, ART_covg, HIV_prev, mort_H0, mort_H1, mdr2015, mdr2022, mdriniTX, sym_sim);
        #%         keyboard;
        
    if out.imag>0 : #if imag(out)>0
        out = math.inf*calfn['sgn'] #out = Inf*calfn.sgn;
        aux = []   
    else:
        #% --- Get additional outputs and package --------------------------
        aux['soln']     = soln# %[soln, t] #aux.soln     = soln; %[soln, t];         % 2003:2023
        aux['allsol']   = np.concatenate([soln2, soln3[1:,:], soln4[1:,:]]) #aux.allsol   = [soln2; soln3(2:end,:); soln4(2:end,:)];  % 1997:2023
        aux['inc2022']  = inc2022 #aux.inc2022  = inc2022;
        aux['inc2015'] = inc2015 #aux.inc2015 = inc2015;
        aux['inc2000'] = inc2000 #aux.inc2015 = inc2015;

        aux['inc_h1']   = inc_h1 #aux.inc_h1   = inc_h1;
        aux['noti']     = noti2022 #aux.noti     = noti;
        aux['noti_cu']  = noti_cu #aux.noti_cu  = noti_cu;
        aux['ART_covg'] = ART_covg#aux.ART_covg = ART_covg;
        aux['HIV_prev'] = HIV_prev #aux.HIV_prev = HIV_prev;
        aux['mort_H02000']  = mort_H02000 #aux['mort_H0']  = mort_H0 #aux.mort_H0  = mort_H0;
        aux['mort_H02022']  = mort_H02022 #aux.mort_H02022  = mort_H02022;
        aux['mort_H1']  = mort_H12022#aux['mort_H1']  = mort_H1#aux.mort_H1  = mort_H1;
        aux['mdr2015']  = mdr2015 #aux.mdr2015  = mdr2015;
        aux['mdr2022']  = mdr2022#aux.mdr2022  = mdr2022;
        aux['mdriniTX'] = mdriniTX #aux.mdriniTX = mdriniTX;
        aux['sym']      = sym_sim #aux.sym      = sym_sim;
        aux['M1']       = M1 #aux.M1       = M1;
    #end
        
    return (out, aux)